# String Routines.
toString = str.__str__

# Todo: use regexprs..?
# Todo: test against regexprs.
def buildNonletters():
    return ''.join(chr(c) for c in range(0, ord('A'))) + \
           ''.join(chr(c) for c in range(ord('Z') + 1, ord('a'))) + \
           ''.join(chr(c) for c in range(ord('z') + 1, 256))

def buildAscii():
    return ''.join(chr(c) for c in range(0, 256))

try: from curses.ascii import isprint as isPrintable
except ImportError:
    _printable_chr = '\' !"#$%&\\\'()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\\\\]^_`abcdefghijklmnopqrstuvwxyz{|}~\''
    def isPrintable(c):
        return c in _printable_chr

def buildNonprintable():
    return ''.join(chr(c) for c in range(0, 256) \
                   if not chr(c) == '\n' and not isPrintable(chr(c)))

_nonletters = buildNonletters()
_ascii = buildAscii()

_nonprintable = buildNonprintable()

_3translate_ascii = dict((n, n) for n in _ascii)
_3translate_ascii_letters = dict(_3translate_ascii)
_3translate_ascii_printable = dict(_3translate_ascii)

for n in _nonletters:
    _3translate_ascii_letters[n] = None
for n in _nonprintable:
    _3translate_ascii_printable[n] = None

def getLetters(string):
    return string.translate(_3translate_ascii_letters)

def getPrintable(string):
    return string.translate(_3translate_ascii_printable)

def splitOne(s, x):
    # Command-line argument passing routine.
    s = s.lstrip()
    parts = s.split(x, 1)
    if not parts:
        return ('', '')

    if len(parts) == 1:
        return (parts[0], '')

    return (parts[0], parts[1])

chopOne = splitOne

def SORP(n):
    return '' if n == 1 else 's'

def ANORA(s):
    return 'an' if s[:1].lower() in 'aeiou' else 'a'


VARIABLE_CODE = '$'
def InterpolateStringVariables(string, **values):
    parts = []
    position = 0

    while True:
        start = string.find(VARIABLE_CODE, position)
        if start < 0:
            break

        end = string.find(VARIABLE_CODE, start + 1)
        if end < 0:
            raise ValueError('Unterminated variable name: %s' % (string[start:]))

        if start > position:
            parts.append(string[position:start])

        name = string[start+1:end]
        if name:
            parts.append(values.get(name, ''))
        else:
            parts.append(VARIABLE_CODE)

        position = end + 1

    if position < len(string):
        parts.append(string[position:])

    return ''.join(parts)

# Other interpolation tools.
import re
SUBMETHOD_PATTERN = re.compile(r'(%(identifier)s)\$(%(identifier)s)\s*\(\s*(\))?' % \
                               dict(identifier = '[a-zA-Z]+[a-zA-Z0-9_]*'))

def TranslateSubmethodPatterns(string):
    def t():
        i = SUBMETHOD_PATTERN.scanner(string)
        c = 0

        while True:
            m = i.search()
            if m is None:
                break

            (s, e) = m.span()
            if s > c:
                yield string[c:s]

            c = e

            (a, b, p) = m.groups()
            if p:
                yield "call(%s, '%s')" % (a, b)
            else:
                yield "call(%s, '%s', " % (a, b)

        if c < len(string):
            yield string[c:]

    return ''.join(t())

# "{percentage(me.hit, me.max_hit)}"

class Expression(str):
    def __call__(self, sandbox):
        # Should, in practice, separate globals, locals
        return sandbox.eval(self)

def TokenizeInterpolatedExpression(string):
    s = 0
    n = len(string)

    while True:
        i = string.find('{', s)
        if i < 0:
            break

        if i > s:
            yield string[s:i]

        # Find ending '}' lexically (looking for nests and quotes).
        v = 0
        q = False

        i += 1
        for e in range(i, n):
            k = string[e]
            if q:
                # Handle quoted matter.
                if k == q:
                    if t != '\\':
                        q = False
                else:
                    q = k

                t = k # always gets set

            elif k == '{':
                v += 1 # deepness
            elif k == '}':
                if v:
                    v -= 1
                else:
                    break
        else:
            raise SyntaxError('No terminating }')

        yield Expression(string[i:e])
        s = e + 1

    if s < n:
        yield string[s:n]

class Interpolated(str):
    @classmethod
    def translateValue(self, sandbox, value):
        def i(t):
            if isinstance(t, Expression):
                return str(t(sandbox))
            if isinstance(t, str):
                return t

            return str(t)

        return ''.join(i(t) for t in TokenizeInterpolatedExpression(value))

    def __call__(self, sandbox):
        return self.translateValue(sandbox, self)

    __mod__ = __call__

class Interpolated2(Interpolated):
    @classmethod
    def translateValue(self, sandbox, value):
        # Convert submethod syntax patterns in string.
        value = TranslateSubmethodPatterns(value)
        return Interpolated.translateValue(sandbox, value)

color_codes = dict(N = 0, r = 31, g = 32, y = 33,
                   b = 34, m = 35, c = 36, w = 37)
 # //#define KNRM  "\033[0m"
 # //#define KRED  "\033[31m"
 # //#define KGRN  "\033[32m"
 # //#define KYEL  "\033[33m"
 # //#define KBLU  "\033[34m"
 # //#define KMAG  "\033[35m"
 # //#define KCYN  "\033[36m"
 # //#define KWHT  "\033[37m"

def parse_color(text):
    SC = 1
    CC = 2

    s = SC
    b = ''

    for c in text:
        if c == '&':
            if s == CC:
                if b:
                    yield b
                    b = ''

                yield '&'
                s = SC
            else:
                s = CC

        elif s == CC:
            if b:
                yield b
                b = ''

            try: yield '\033[%sm' % color_codes[c]
            except KeyError:
                yield '&' + c

            s = SC
        else:
            b += c

    if b:
        yield b


def renderTemplate(*args, **kwd):
    global renderTemplate
    from stuphos.runtime.architecture.api import renderTemplate
    return renderTemplate(*args, **kwd)


def compactify(string, width):
    string = str(string)
    if len(string) <= width:
        return string

    return '%s...%s' % (string[:int(width/2)-1], string[-(int(width/2)-2):])


def escapeXmlOutput(s):
    # Todo: regexprs
    s = s.replace('&', '&amp;')
    s = s.replace('<', '&lt;')
    s = s.replace('>', '&gt;')
    s = getPrintable(s)
    return s

# escapeStreamOutput = escapeXmlOutput
# escapeStreamOutput = getPrintable

def escapeOutputIdentity(s):
    return s


def safeAttr(s):
    'Return string escaped for embedding in an HTML attribute value.'

    return s.replace('"', '&quot;')

def escapeSingleQuotes(s):
    return s.replace("'", "\\'")
def escapeDoubleQuotes(s):
    return s.replace('"', '\\"')

def escapeQuotes(s):
    return escapeSingleQuotes(escapeDoubleQuotes(s))

def escapeHtmlJSQuotes(s):
    return safeAttr(escapeSingleQuotes(s))


def nling(function):
    def nl(*args, **kwd):
        return '\n'.join(function(*args, **kwd))

    return nl

def indent(string, tab = '    ', level = 1):
    tab *= level
    return tab + ('\n' + tab).join(string.split('\n'))


def isYesValue(value):
    if value is True:
        return True
    if isinstance(value, str):
        return value.lower() in ['yes', 'true', '1', 'on']

    return bool(value)

def isNoValue(value):
    if value is False:
        return True
    if isinstance(value, str):
        return value.lower() in ['no', 'false', '0', 'off']

    return not bool(value)



# Frozen because it's being merged back into main.
# # Originally from GardenSnake:
# #
# #          Andrew Dalke / Dalke Scientific Software, LLC
# #             30 August 2006 / Cape Town, South Africa

# # Differences:
# #  - utilization of xrange instead of range
# #  - import ply.lex explicitly here
# #  - added module = keyword to lexer constructor to separate from this module.

# ## I implemented INDENT / DEDENT generation as a post-processing filter

# # The original lex token stream contains WS and NEWLINE characters.
# # WS will only occur before any other tokens on a line.

# # I have three filters.  One tags tokens by adding two attributes.
# # "must_indent" is True if the token must be indented from the
# # previous code.  The other is "at_line_start" which is True for WS
# # and the first non-WS/non-NEWLINE on a line.  It flags the check so
# # see if the new line has changed indication level.

# # Python's syntax has three INDENT states
# #  0) no colon hence no need to indent
# #  1) "if 1: go()" - simple statements have a COLON but no need for an indent
# #  2) "if 1:\n  go()" - complex statements have a COLON NEWLINE and must indent
# NO_INDENT = 0
# MAY_INDENT = 1
# MUST_INDENT = 2

# def countws(n):
#     w = 0
#     for c in n:
#         if c in '\t ':
#             w += 1
#         else:
#             break

#     return w

# # only care about whitespace at the start of a line
# def track_tokens_filter(lexer, _tokens):
#     # Todo: this might be a good place to insert lineno info for each token.
#     lexer.at_line_start = at_line_start = True
#     indent = NO_INDENT
#     first = False # True # todo: consider preceding whitespace
#     # saw_colon = False

#     lookahead = [None]

#     def readYaml():
#         l = lexer # .lexer
#         i = l.lexpos
#         d = l.lexdata
#         n = len(d)

#         def line(i):
#             b = []
#             while i < n:
#                 c = d[i]
#                 i += 1
#                 if c == '\n':
#                     break

#                 b.append(c)

#             return (i, ''.join(b))

#         f = True
#         y = []

#         while i < n:
#             s = i
#             (i, b) = line(i)
#             w = countws(b)

#             if w == len(b):
#                 if b or not f:
#                     y.append(b[ws:])

#                 continue

#             if f:
#                 f = False
#                 ws = w

#             elif w < ws:
#                 l.lexpos = s # (-1 doesn't seem to matter: or does it?)
#                 break

#             y.append(b[ws:])
#         else:
#             l.lexpos = n # (-1 doesn't seem to matter: or does it?)

#         return '\n'.join(y) + '\n'


#     def checkYaml(name):
#         n = next(tokens) # or _tokens?
#         if n.type == 'COLON':
#             yaml = readYaml()
#             token = _new_token('YAML', name.lineno) # todo: rename type to something like 'BLOCK'
#             token.name = name
#             token.value = yaml
#             token.at_line_start = at_line_start # probably always true!
#             token.must_indent = (indent == MUST_INDENT)

#             # print repr(lexer.lexdata[lexer.lexpos:])
#             return token

#         lookahead[0] = n

#     @apply
#     def tokens():
#         i = iter(_tokens)
#         while True:
#             if lookahead[0]:
#                 yield lookahead[0]
#                 lookahead[0] = None
#                 continue

#             yield next(i)

#     for token in tokens:
#         token.at_line_start = at_line_start

#         # Experimental inline yaml support:
#         if (first or at_line_start) and token.type == 'NAME':
#             yaml = checkYaml(token)
#             if yaml:
#                 indent = NO_INDENT
#                 yield yaml
#                 continue

#         if token.type == "COLON":
#             at_line_start = False
#             indent = MAY_INDENT
#             token.must_indent = False
            
#         elif token.type == "NEWLINE":
#             at_line_start = True
#             if indent == MAY_INDENT:
#                 indent = MUST_INDENT
#             token.must_indent = False

#         elif token.type == "WS":
#             # XXXXX Why doesn't this work anymore/ever did work??
#             # assert token.at_line_start == True
#             # at_line_start = True # would already be true...
#             token.must_indent = False

#         else:
#             # A real token; only indent after COLON NEWLINE
#             if indent == MUST_INDENT:
#                 token.must_indent = True
#             else:
#                 token.must_indent = False
#             at_line_start = False
#             indent = NO_INDENT

#         yield token
#         lexer.at_line_start = at_line_start

# def _new_token(type, lineno):
#     tok = lex.LexToken()
#     tok.type = type
#     tok.value = None
#     tok.lineno = lineno
#     tok.lexpos = 0
#     return tok

# # Synthesize a DEDENT tag
# def DEDENT(lineno):
#     return _new_token("DEDENT", lineno)

# # Synthesize an INDENT tag
# def INDENT(lineno):
#     return _new_token("INDENT", lineno)


# # Track the indentation level and emit the right INDENT / DEDENT events.
# def indentation_filter(tokens):
#     # A stack of indentation levels; will never pop item 0
#     levels = [0]
#     token = None
#     depth = 0
#     prev_was_ws = False
#     for token in tokens:
#         # print depth, levels, token

#         ##    if 1:
#         ##        print "Process", token,
#         ##        if token.at_line_start:
#         ##            print "at_line_start",
#         ##        if token.must_indent:
#         ##            print "must_indent",
#         ##        print
                
#         # WS only occurs at the start of the line
#         # ???
#         # There may be WS followed by NEWLINE so
#         # only track the depth here.  Don't indent/dedent
#         # until there's something real.
#         if token.type == "WS":
#             # assert depth == 0 # wtf line start
#             # yeah, but this may be needed to properly emit indentation tokens
#             # ALTHOUGH, it doesn't seem to effect _current_ absence.
#             depth = len(token.value)
#             prev_was_ws = True
#             # WS tokens are never passed to the parser
#             continue

#         if token.type == "NEWLINE":
#             depth = 0
#             if prev_was_ws or token.at_line_start:
#                 # ignore blank lines
#                 # XXX
#                 continue
#             # pass the other cases on through
#             yield token
#             continue

#         prev_was_ws = False # assumed to be in right location
#         if token.type == 'YAML':
#             # supposed to simulate the line-start below, but does
#             # this work in complex nests?
#             if depth in levels:
#                 levels.remove(depth)

#             depth = 0 # err what if must_indent is reset?!?

#             if token.must_indent:
#                 yield INDENT(token.lineno)

#             yield token

#             if token.must_indent:
#                 yield DEDENT(token.lineno)

#             continue

#         # then it must be a real token (not WS, not NEWLINE)
#         # which can affect the indentation level

#         if token.must_indent:
#             # The current depth must be larger than the previous level
#             if not (depth > levels[-1]):
#                 raise IndentationError("expected an indented block")

#             levels.append(depth)
#             yield INDENT(token.lineno)

#         elif token.at_line_start:
#             # Must be on the same level or one of the previous levels
#             if depth == levels[-1]:
#                 # At the same level
#                 pass
#             elif depth > levels[-1]:
#                 raise IndentationError("indentation increase but not in new block")
#             else:
#                 # Back up; but only if it matches a previous level
#                 try:
#                     i = levels.index(depth)
#                 except ValueError:
#                     raise IndentationError("inconsistent indentation")
#                 for _ in xrange(i+1, len(levels)):
#                     yield DEDENT(token.lineno)
#                     levels.pop()

#         yield token

#     ### Finished processing ###

#     # Must dedent any remaining levels
#     if len(levels) > 1:
#         assert token is not None
#         for _ in xrange(1, len(levels)):
#             yield DEDENT(token.lineno)
    

# # The top-level filter adds an ENDMARKER, if requested.
# # Python's grammar uses it.
# def filter(lexer, add_endmarker = True):
#     token = None
#     tokens = iter(lexer.token, None)
#     tokens = track_tokens_filter(lexer, tokens)
#     for token in indentation_filter(tokens):
#         yield token

#     if add_endmarker:
#         lineno = 1
#         if token is not None:
#             lineno = token.lineno
#         yield _new_token("ENDMARKER", lineno)

# # Combine Ply and my filters into a new lexer
# import ply.lex as lex

# class IndentLexer(object):
#     def __init__(self, module=None, debug=0, optimize=0, lextab='lextab', reflags=0):
#         self.lexer = lex.lex(module=module, debug=debug, optimize=optimize, lextab=lextab, reflags=reflags)
#         self.token_stream = None
#         self.lineno = 1
#     def input(self, s, add_endmarker=True):
#         self.lexer.paren_count = 0
#         self.lexer.input(s)
#         self.token_stream = filter(self.lexer, add_endmarker)
#     def token(self):
#         try:
#             return self.token_stream.next()
#         except StopIteration:
#             return None



'lib/etc/players'
__all__=['PLRFILE', 'PLRDBIntoSQL', 'DEFAULT_PLRFL', 'DEFAULT_STORE']

from struct import *
from os.path import exists

# C-Struct Format.
MAX_NAME_LENGTH=20
MAX_EMAIL_LENGTH=80
MAX_PWD_LENGTH=10
MAX_AFFECT=32
MAX_COLOR_SET=32
HOST_LENGTH=40
MAX_SKILLS=200
MAX_TONGUE=3

char_specials_fmt='i3l5h\n'
player_specials_fmt='%ds\n' % (calcsize('b') * MAX_SKILLS+1)
player_specials_fmt+='x\n'
player_specials_fmt+='%ds\n' % (calcsize('i') * MAX_TONGUE)
player_specials_fmt+='ibhI3lB3b3il3i\n'
abilities_fmt='7b\n'
points_fmt='12h3i2b\n'

affected_fmt='2h2bl\n'
affected_fmt+='%dx' % calcsize('P') # padd out next pointer

ending_padding='' # '1191x'

pfilel_fmt='%ds %ds 6i3bh2i2B %ds\n' % (MAX_NAME_LENGTH+1, MAX_EMAIL_LENGTH+1, MAX_PWD_LENGTH+1)
pfilel_fmt+=char_specials_fmt +player_specials_fmt +abilities_fmt +points_fmt
pfilel_fmt+='%ds' % (calcsize(affected_fmt) * MAX_AFFECT)
pfilel_fmt+='%ds' % (calcsize('i') * MAX_COLOR_SET)
pfilel_fmt+='2i'
pfilel_fmt+='%ds' % (HOST_LENGTH+1)
pfilel_fmt+=ending_padding

# Names.
saving_throw_names=('savingthrow1', 'savingthrow2', 'savingthrow3', 'savingthrow4', 'savingthrow5')

char_specials_names=('alignment', 'idnum', 'act', 'affectedby') +saving_throw_names
player_specials_names=('skills', 'talks', 'wimp', 'freeze', 'invis', 'loadroom',
                'pref', 'pref2', 'disp', 'badpws', 'drunk', 'hunger', 'thirst',
                'spells2learn', 'maritalstatus', 'num_marriages', 'spouse',
                'spellattack1', 'spellattack2', 'remortpoints')

abilities_names=('str', 'stradd', 'intel', 'wis', 'dex', 'con', 'cha')
points_names=('mana', 'maxmana', 'hit','maxhit', 'move', 'maxmove', 'deaths', 'mkills', 'pkills',
        'dts', 'qpoints', 'armor', 'gold', 'bank', 'exp', 'hitroll', 'damroll')

pfilel_names=('name', 'email', 'remort', 'race', 'clan', 'clanrank', 'pagelength', 'saveroom',
        'sex', 'chclass', 'level', 'hometown', 'birth', 'played', 'weight', 'height', 'pwd')\
        +char_specials_names +player_specials_names +abilities_names +points_names\
        +('affected', 'color', 'lastlogin', 'lastlogon', 'host')

pfilel_ncount=len(pfilel_names)
pfilel_types=['']*pfilel_ncount

# from structfmt import stfmti

#for n in 'pfilel_fmt', 'char_specials_fmt', 'player_specials_fmt', 'abilities_fmt', 'points_fmt', 'affected_fmt':
#   print n, `vars()[n]`, calcsize(vars()[n]), len(list(stfmti(vars()[n])))

#for n in 'pfilel_names', 'saving_throw_names', 'char_specials_names', 'player_specials_names', 'abilities_names', 'points_names':
#   print n, `vars()[n]`, len(vars()[n])

pfilel_size=calcsize(pfilel_fmt)
pfilel_sqldecl=', '.join([' '.join(p) for p in zip(pfilel_names, pfilel_types)])

# General Routines.
# ToDo: incorporate the original pfile-position

# Because of packing in the server code, the char_file_u struct
# does exactly match what we would expect to read with python
# struct.  This is a version-level hack to align properly.
ALIGNMENT_HACK_LEN = 1
ALIGNMENT_HACK_BUF = '\x00' * ALIGNMENT_HACK_LEN

def _read_player_buffer(playerfl):
    'Read one buffer capable of being unpacked a la pfilel_fmt.'

    b = playerfl.read(pfilel_size - ALIGNMENT_HACK_LEN)
    if b:
        return b + ALIGNMENT_HACK_BUF

    raise EOFError

def goto_pfilepos(fl, pfilepos):
    return fl.seek(pfilepos*pfilel_size)

def read_player(playerfl, pfilepos = None):
    if pfilepos is not None:
        goto_pfilepos(playerfl, pfilepos)

    return unpack(pfilel_fmt, _read_player_buffer(playerfl))

def write_player(plrrecord, playerfl, pfilepos = None):
    if pfilepos is not None:
        goto_pfilepos(playerfl, pfilepos)

    return playerfile.write(pack(pfilel_fmt, plrrecord))

class PlayerFileReader(object):
    "Iterate over load_source(filename)"

    def __init__(self, source):
        self.source = source

    def __repr__(self):
        return 'players: '+repr(self.source)
    def __iter__(self):
        return self.read_player_entries()

    def openPlayerFileSource(self, source):
        if type(source) is file:
            return file

        return open(source)

    def read_player_entries(self):
        'Generator -- Streams player records'
        plrfile = self.openPlayerFileSource(self.source)

        try:
            while True:
                yield unpack(pfilel_fmt, _read_player_buffer(plrfile))

        except EOFError:
            pass

## Aliases:
PLRFILEReader=PLRFILE=PlayerFileReader

def plr2dict(player):
    return dict((pfilel_names[i], player[i]) for i in range(len(pfilel_names)))
def dict2plr(dict):
    return tuple(dict[n] for n in pfilel_names)

def cstring(string):
    i = string.find('\x00')
    if i >= 0:
        return string[:i]

    return string

pfilel_conv = dict(name  = cstring,
                   host  = cstring,
                   email = cstring,
                   pwd   = cstring)

class Player:
    def __init__(self, record):
        # Generate tuple.
        if type(record) is file:
            record = read_player(record)
        elif type(record) is str:
            record = unpack(pfilel_fmt, record)

        # Generate dict.
        if type(record) is dict:
            record = record.copy()
        elif type(record) in (list, tuple):
            record = plr2dict(record)
        else:
            raise TypeError(type(record).__name__)

        # Convert.
        for (name, conv) in pfilel_conv.items():
            record[name] = conv(record[name])

        self.__dict__ = record

    def __iter__(self):
        gi = self.__dict__.__getitem__
        return (gi(n) for n in pfilel_names)

    def __repr__(self):
        return '<%s: %r [#%d]>' % (self.__class__.__name__,
                                   self.name, self.idnum)

DEFAULT_PFILE = 'lib/etc/players'

class PlayerIndex:
    def __init__(self, source, FileReaderClass = PlayerFileReader,
                 PlayerClass = Player):

        self.byName = dict()
        self.byIdnum = dict()

        for record in FileReaderClass(source):
            player = PlayerClass(record)
            self.byName[player.name] = player
            self.byIdnum[player.idnum] = player

    def __repr__(self):
        return '<%s (%d players)>' % (self.__class__.__name__, len(self))

    def __getitem__(self, item):
        if type(item) is int:
            return self.byIdnum[item]

        return self.byName[item]

    def __len__(self):
        return len(self.byName)
    def __iter__(self):
        return iter(self.byName.values())

    def idnums(self):
        return list(self.byIdnum.keys())
    def names(self):
        return list(self.byName.keys())

# Strings & Aliases.
def ReadStringsFile(file):
    from .dblib import FileReader
    reader = FileReader(file)

    return dict(title = reader.tildeString('title'),
                prename = reader.tildeString('prename'),
                wizname = reader.tildeString('wizname'),
                poofin = reader.tildeString('poofin'),
                poofout = reader.tildeString('poofout'),
                description = reader.tildeString('description'),
                plan = reader.tildeString('plan'))

class Alias:
    def __init__(self, name, replacement, complex = False):
        self.name = name
        self.replacement = replacement
        self.complex = bool(complex)

    def __repr__(self):
        return '<%s%s %s = %r>' % ('(complex)' if self.complex else '',
                                   self.name, self.replacement)

def ReadAliasFile(file, map = False):
    from .dblib import FileReader
    reader = FileReader(file)

    def _():
        # Generate alias structures.
        try:
            while True:
                yield Alias(reader.tildeString('alias name'),
                            reader.tildeString('alias replacement'),
                            (reader.readLine() == '1'))

        except EOFError:
            pass

    # Return an associative mapping keyed on alias name, or just the iterator.
    return dict((a.name, a) for a in _()) if map else _()

class PlrFileSupplemental:
    def __init__(self, name):
        self.name = name
        self.load()

    def __repr__(self):
        return '<%s for %s>' % (self.__class__.__name__, self.name)

    @property
    def filename(self):
        from stuphlib import directory as stuphdir
        return stuphdir.StuphLIB['PLAYER[%s]:%s' % (self.TOPIC, self.name)]

    def load(self):
        raise NotImplementedError
    def save(self):
        raise NotImplementedError

class Strings(PlrFileSupplemental):
    TOPIC = 'STRINGS'

    def load(self):
        self.strings = ReadStringsFile(self.filename)
        self.__getitem__ = self.strings.__getitem__
        self.__setitem__ = self.strings.__setitem__
        self.keys = self.strings.keys

class Aliases(PlrFileSupplemental):
    TOPIC = 'ALIAS'

    def load(self):
        self.aliases = ReadAliasFile(self.filename, map = True)
        self.__getitem__ = self.aliases.__getitem__
        self.keys = self.aliases.keys


# SQLITE
NEW_PLAYERS_TABLE='CREATE TABLE players(%s)' % ', '.join(pfilel_names)
NEW_PLAYER='INSERT OR REPLACE INTO players('+pfilel_sqldecl+') VALUES('+', '.join(['?']*pfilel_ncount)+')'

#pfilel_types=('varchar(%d)'%MAX_NAME_LENGTH+1, 'varchar(%d)'%MAX_EMAIL_LENGTH+1)+('integer',)*6

class PLRDBIntoSQL(PlayerFileReader):
    "Plugs a PLRFILE into an SQLite database"

    def __init__(self, store=None, loadsource=None):
        # PlayerFileReader.__init__(self)
        self.open_store(type(store) is str and (store,) or store)
        if loadsource:
            self.save_players(self.read_player_entries(loadsource))

    def write_player(self, player, pfilepos = 0):
        'Install player into cursory connection'
        # from pprint import pprint as pp
        # pp(player)
        # raw_input()
        # print 'len(player[1])', len(player[1])

        self.x(NEW_PLAYER, player)
        # self.x('UPDATE players SET pfilepos=%d WHERE name=%r' % (pfilepos, player[1][0]))

    def save_players(self, players = None):
        "Stream all players given by the `players' iterable through write_player -- Also creates self.stmts and self.players."
        plrs=self.players=[]
        append=plrs.append

        stmts=self.stmts=[]
        append2=stmts.append

        pfilepos=0
        for p in players:
            append2(p[0])
            append(p[1])

            self.write_player(p, pfilepos)
            pfilepos+=1

        print('Saved', len(plrs), 'Players ...')

    def open_store(self, store):
        e=not exists(store[0])

        from pysqlite2 import dbapi2 as sqlite
        c=self.connection=sqlite.connect(*store)
        u=self.u=c.cursor()
        x=self.x=u.execute

        if e:
            x(NEW_PLAYERS_TABLE)
            x('ALTER TABLE players ADD (pfilepos)')

    def commit(self):
        self.connection.commit()

DEFAULT_PLRFL='lib/etc/players'
DEFAULT_STORE='lib/etc/players.db' # ':memory:'

def loadSql():
    from psyco import full;full()

    plrdb=PLRDBIntoSQL(DEFAULT_STORE)
    plrdb.save_players(plrdb.read_player_entries(DEFAULT_PLRFL))
    plrdb.commit()

    u=plrdb.u
    x=plrdb.x
    a=lambda:x('select * from players')

# Test Bed.
def read_player_analysis(playerfl):
    try:
        b=playerfl.read(pfilel_size - 1)
        if b == '':
            raise EOFError

        return unpack(pfilel_fmt, b + '\x00')

    except error as e:
        print('read/record: %d / %d' % (len(b), pfilel_size))
        raise e

def analyze_player_file(plrfile):
    nr_records = 0

    import os
    plrfile.seek(0, os.SEEK_END)
    bytes = plrfile.tell()
    print('size of player file:', bytes)
    print('number of entries:', float(bytes) / pfilel_size)
    plrfile.seek(0)

    try:
        while True:
            plr = read_player_analysis(plrfile)
            nr_records += 1

    except EOFError:
        pass
    except KeyboardInterrupt:
        pass
    finally:
        print('number of records:', nr_records)

def readWholeFile(filename):
    from io import StringIO
    buf = StringIO()

    import os
    goal = os.stat(filename).st_size

    file = open(filename)
    size = 0
    while size < goal:
        b = file.read(512)
        if b == '':
            print('null read!')

        buf.write(b)
        size += len(b)

    buf.seek(0)
    print('read into buffer:', file.tell())
    return buf

# XML Generation.
def getTab(kwd):
    return kwd.pop('_tab', '  ') * kwd.pop('_indent', 0)

def getAttr(nv):
    return '%s="%s"' % nv

def xmlOpen(tag, **kwd):
    tab = getTab(kwd)
    close = kwd.pop('_close', False)
    attrs = ' '.join(getAttr(nv) for nv in kwd.items())
    return '%s<%s%s%s%s>' % (tab, tag, ' ' if attrs else '', attrs,
                             ' /' if close else '')

def xmlClose(tag, **kwd):
    return '%s</%s>' % (getTab(kwd), tag)

import base64
def base64_encode(value, **kwd):
    value = base64.encodestring(value)
    tab = getTab(kwd)
    if tab:
        value = tab + ('\n' + tab).join(value.split('\n'))

    return value

XML_SKIP = ['name', 'idnum']
XML_LONG_FORM = ['color', 'affected', 'skills', 'talks']
XML_ENCODING = dict(color    = ('base64', base64_encode),
                    affected = ('base64', base64_encode),
                    skills   = ('base64', base64_encode),
                    talks    = ('base64', base64_encode))

def generatePlayerFileXML(plrfile):
    index = PlayerIndex(plrfile)

    yield '<?xml version="1.0">'
    yield xmlOpen('playerIndex')
    for plr in index:
        yield xmlOpen('player', name = plr.name, idnum = plr.idnum, _indent = 1)
        for attr in pfilel_names:
            if attr in XML_SKIP:
                continue

            value = getattr(plr, attr)
            encoding = XML_ENCODING.get(attr)
            if encoding is not None:
                (encoding, encode) = encoding
                value = encode(value, _indent = 3)

            if attr in XML_LONG_FORM:
                if encoding:
                    opening = xmlOpen(attr, encoding = encoding, _indent = 2)
                else:
                    opening = xmlOpen(attr, _indent = 2)

                if type(value) in (str, str):
                    newline = value.rfind('\n')
                    if newline < 0:
                        yield '%s%s%s' % (opening, value, xmlClose(attr))
                    else:
                        yield '%s\n%s%s%s' % (opening, value,
                                              '\n' if newline != (len(value) - 1)
                                                else '',
                                              xmlClose(attr, _indent = 2))

            else:
                yield xmlOpen(attr, value = value, _indent = 2, _close = True)

        yield xmlClose('player', _indent = 1)

    yield xmlClose('playerIndex')

def reportPlayerFileXML(plrfile):
    for line in generatePlayerFileXML(plrfile):
        print(line)

class JSONPlayerFileReader(PlayerFileReader):
    # Attempts to verify that the player record is jsonifiable.
    # It does this by encoding the following fields:
    ENCODE_FIELDS = ['color', 'affected', 'skills', 'talks']
    CSTRIP_FIELDS = ['name', 'email', 'pwd', 'host']

    # If it finds that it cannot jsonify the result, it skips
    # the record (because it is malformed anyway).
    def read_player_entries(self):
        from sys import stderr
        nr = 0
        for record in super(self.__class__, self).read_player_entries():
            # Make mutable:
            record = list(record)
            nr += 1

            for field in self.ENCODE_FIELDS:
                i = pfilel_names.index(field)
                record[i] = base64_encode(record[i])

            # what about pfilel_conv?
            for field in self.CSTRIP_FIELDS:
                i = pfilel_names.index(field)
                record[i] = cstring(record[i])

            invalid = self.json_invalid(record)
            if invalid:
                print('record #%d:' % nr, file=stderr)
                for i in range(len(pfilel_names)):
                    invalid = self.json_invalid(record[i])
                    if invalid:
                        print('  %s: %s' % (pfilel_names[i], invalid), file=stderr)
            else:
                # Make immutable, again:
                yield tuple(record)

    try: from django.utils.simplejson import dumps as _json_dumps
    except ImportError:
        try: from json import dumps as _json_dumps
        except ImportError:
            from simplejson import dumps as _json_dumps

    _json_dumps = staticmethod(_json_dumps)

    def json_invalid(self, object):
        try: self._json_dumps(object)
        except Exception as e: return str(e)
        else: return False

def reportPlayerFileJSON(plrfile, stream = False):
    # from simplejson import dump
    from json import dump
    from sys import stdout

    # One big record:
    players = PlayerIndex(plrfile, FileReaderClass = JSONPlayerFileReader)

    if stream:
        for plr in players:
            dump(list(plr), stdout)
    else:
        # players = tuple(tuple(plr) for plr in players)
        players = list(map(list, players))

        dump(players, stdout)

def report_player_file(plrfile, options):
    index = PlayerIndex(plrfile)
    print(index)
    for plr in index:
        print('%-15s [%s]' % (plr.name, plr.email))

def parse_cmdln(argv = None):
    from optparse import OptionParser
    parser = OptionParser()
    parser.add_option('-r', '--report', action = 'store_true')
    parser.add_option('-x', '--xml', action = 'store_true')
    parser.add_option('-j', '--json', action = 'store_true')
    parser.add_option('-s', '--stream', action = 'store_true')

    parser.add_option('-f', '--filename')

    parser.add_option('-g', '--debug', action = 'store_true')
    parser.add_option('-b', '--buffer', action = 'store_true')
    return parser.parse_args()

def main(argv = None):
    from pdb import runcall
    (options, args) = parse_cmdln(argv)

    plrfile = options.filename
    if not plrfile:
        plrfile = args[0] if args else r'H:\StuphMUD\lib\etc\players'

    # Just report what we can.
    if options.report:
        if options.xml:
            reportPlayerFileXML(plrfile)
        elif options.json:
            if options.debug:
                runcall(reportPlayerFileJSON, plrfile, options.stream)
            else:
                reportPlayerFileJSON(plrfile, options.stream)
        else:
            report_player_file(plrfile)
    else:
        # Read entire file into buffer before analysis. (Pythonwin)
        if options.buffer:
            plrfile = readWholeFile(plrfile)
        else:
            plrfile = open(plrfile)

        # Perform analysis read. (careful)
        if options.debug:
            runcall(analyze_player_file, plrfile)
        else:
            analyze_player_file(plrfile)

if __name__=='__main__':
    main()

'''
AppEngine:                 AE           Ormus        Difference
21s 81s 6i3bh2i2B 11s    : 157          157
i3l5h                    : 42           26           > 16
201s                     : 201          201
x                        : 1            1
12s                      : 12           12
ibhI3lB3b3il3i           : 76           56           > 20
7b                       : 7            7
12h3i2b                  : 38           38
768s128s2i41s            : 945          689          > 256

                                                     > 292 bytes

Ormus:
21s 81s 6i3bh2i2B 11s    : 157
i3l5h                    : 26
201s                     : 201
x                        : 1
12s                      : 12
ibhI3lB3b3il3i           : 56
7b                       : 7
12h3i2b                  : 38
512s128s2i41s            : 689
'''



'This package stores SQLObject schema in separately-imported modules.'
from os.path import abspath

IN_MEMORY = ':memory:'
DEFAULT_PROTOCOL = 'sqlite'

def _importMember(module):
	parts  = module.split('.')
	module = __import__('.'.join(parts[:-1]))

	for m in parts[1:]:
		module = getattr(module, m)

	# Or member
	return module

def openProcessConnection(path, protocol = DEFAULT_PROTOCOL):
	if path == IN_MEMORY:
		path = '/' + IN_MEMORY
	else:
		path = abspath(path)

	c = sqlobject.connectionForURI('%s:%s' % (protocol, path))
	sqlobject.sqlhub.processConnection = c
	return c

connect = openProcessConnection

def openDatabaseTables(path, *tables, **kwd):
	protocol       = kwd.get('protocol', 'sqlite')
	schema_package = kwd.get('schema_package')

	openProcessConnection(path, protocol)

	def open_table(t):
		if schema_package:
			t = schema_package + '.' + t

		# import sqlobject.classregistry
		# reload(sqlobject.classregistry)

		t = _importMember(t)
		t.createTable(ifNotExists = True)
		return t

	return list(map(open_table, tables))

tables = openDatabaseTables

def getStuphZoneRecordIDByZoneVnum(vnum, cache = {}):
	# Patches up cross-module foreign-key joins.
	if vnum in cache:
		return cache[vnum]

	from .zone import StuphZoneRecord as Z

	r = list(Z.selectBy(vnum = vnum)[:1])

	if len(r) > 0:
		r = r[0].id
		cache[vnum] = r
		return r

	# XXX return -1 ??

# def new_database_file(format = 'stuph-world-%d.db', root_path = 'lib/etc'):
#	# regex = r'stuph-world-(\d+)\.db'
#	from itertools import count
#	from os.path import exists, join
#
#	for dbno in count(1):
#		dbpath = join(root_path, format % dbno)
#		if not exists(dbpath):
#			return dbpath
#
#		assert dbno < 9999

try: import sqlobject
except ImportError:
        class sqlobject:
                pass

        class DisabledException(Exception):
                pass

        def disabled(*args, **kwd):
                raise DisabledException

        connect = openProcessConnection = disabled
        tables = openDatabaseTables = disabled
        
        getStuphZoneRecordIDByZoneVnum = disabled



# This should go in circlemud.world.

'''
[Services]
facility.zone-reset = stuphmud.server.zones.controller.ZoneReset

'''

from stuphos.runtime.facilities import Facility
from stuphos.runtime.architecture import newComponent

class ZoneReset(Facility):
    NAME = 'StuphMUD::Zone::Controller'

    @classmethod
    def create(self):
        return newComponent(self)

    # Enable this as a --reset-world alternative.
    # def onResetComplete(self, ctlr):
    #     self.resetWorld(core)

    def resetWorld(self, world, forceAll = False, bridgeApi = True):
        def zreset(zone):
            return self.resetZone(world, zone, force = forceAll,
                                  bridgeApi = bridgeApi)

        from world import zone, iterent
        iterent(zone, callback = zreset,
                map_results = False)

    def shouldReset(self, zone):
        pass

    def resetZone(self, world, zone, force = False, bridgeApi = True):
        if force or self.shouldReset(zone):
            # debugOn()
            if bridgeApi:
                # print(f'[resetZone] {StuphMUD.StartZoneReset}')
                StuphMUD.StartZoneReset(zone)

            state = (False, None) # last command, last object
            for (nr, cmd) in enumerate(zone.zcommands):
                if cmd.isConditional and not state[0]:
                    continue

                code = cmd.commandCode
                try:
                    if code == 'M':
                        state = self.loadMobile(world, zone, state, nr, cmd)
                    elif code == 'O':
                        state = self.loadObject(world, zone, state, nr, cmd)
                    elif code == 'E':
                        state = self.equipMobile(world, zone, state, nr, cmd)
                    elif code == 'P':
                        state = self.putObject(world, zone, state, nr, cmd)
                    elif code == 'D':
                        state = self.doorState(world, zone, state, nr, cmd)
                    elif code == 'R':
                        state = self.removeObject(world, zone, state, nr, cmd)
                    elif code == 'G':
                        state = self.giveObject(world, zone, state, nr, cmd)

                except Exception as e:
                    print(f'[Reset Zone #{zone.vnum}:{nr} ({code})] {e.__class__.__name__}: {e}')
                    state = (False, None)

            if bridgeApi:
                StuphMUD.CompleteZoneReset(zone)


    # Zone commands.
    # class command:
    #     arg1 = arg2 = arg3 = None
    #     command = commandName = '?'
    #     commandCode = type = '*'
    #     owner = location = container = None
    #     door = doorState = None
    #     mobile = object = room = None

    #     max = maxExisting = 0
    #     line = ''

    #     isConditional = False

    def loadMobile(self, core, zone, state, nr, cmd):
        if core and core.cmdln['options'].verbose > 1:
            print(f'[Zone Reset {zone.name} #{nr}] Load Mobile #{cmd.arg1} in room #{cmd.arg3}, {cmd.arg2} max')

        from world import mobile, mobile_instance, room
        vnum = int(cmd.arg1)

        # debugOn()
        if len(mobile_instance.instances.get(vnum, [])) < int(cmd.arg2):
            proto = mobile.lookup(vnum)
            room = room.lookup(int(cmd.arg3))

            mob = proto.instantiate(room)
            StuphMUD.CreateMobileInstance(mob, room, 'zone')

            return (True, mob)

        return (False, None)

    def loadObject(self, world, zone, state, nr, cmd):
     #    case 'O':         /* read an object */
     #      if (obj_index[ZCMD.arg1].number < ZCMD.arg2) {
        # if (ZCMD.arg3 != NOWHERE) {
        #   obj = read_object(ZCMD.arg1, REAL);
        #   obj_to_room(obj, ZCMD.arg3);
        #   last_cmd = 1;

     #    // zone load item

        # } else {
        #   obj = read_object(ZCMD.arg1, REAL);
        #   IN_ROOM(obj) = NOWHERE;
        #   last_cmd = 1;

     #    // zone load item

        # }
     #      } else
        # last_cmd = 0;
     #      break;
        return (False, None)

    def equipMobile(self, world, zone, state, nr, cmd):
     #    case 'E':         /* object to equipment list */
     #      if (!mob) {
        # ZONE_ERROR("trying to equip non-existant mob, command disabled");
        # ZCMD.command = '*';
        # break;
     #      }
     #      if (obj_index[ZCMD.arg1].number < ZCMD.arg2) {
        # if (ZCMD.arg3 < 0 || ZCMD.arg3 >= NUM_WEARS) {
        #   ZONE_ERROR("invalid equipment pos number");
        # } else {
        #   obj = read_object(ZCMD.arg1, REAL);
        #   equip_char(mob, obj, ZCMD.arg3);
        #   last_cmd = 1;

     #    // zone load item

        # }
     #      } else
        # last_cmd = 0;
     #      break;
        return (False, None)

    def putObject(self, world, zone, state, nr, cmd):
     #    case 'P':         /* object to object */
     #      if (obj_index[ZCMD.arg1].number < ZCMD.arg2) {
        # obj = read_object(ZCMD.arg1, REAL);
        # if (!(obj_to = get_obj_num(ZCMD.arg3))) {
        #   ZONE_ERROR("target obj not found, command disabled");
        #   ZCMD.command = '*';
        #   break;
        # }
        # obj_to_obj(obj, obj_to);
        # last_cmd = 1;

     #  // zone load item

     #      } else
        # last_cmd = 0;
     #      break;
        return (False, None)

    def doorState(self, world, zone, state, nr, cmd):
     #    case 'D':         /* set state of door */
     #      if (ZCMD.arg2 < 0 || ZCMD.arg2 >= NUM_OF_DIRS ||
        #   (world[ZCMD.arg1].dir_option[ZCMD.arg2] == NULL)) {
        # ZONE_ERROR("door does not exist, command disabled");
        # ZCMD.command = '*';
     #      } else
        # switch (ZCMD.arg3) {
        # case 0:
        #   REMOVE_BIT(world[ZCMD.arg1].dir_option[ZCMD.arg2]->exit_info,
        #        EX_LOCKED);
        #   REMOVE_BIT(world[ZCMD.arg1].dir_option[ZCMD.arg2]->exit_info,
        #        EX_CLOSED);
        #   break;
        # case 1:
        #   SET_BIT(world[ZCMD.arg1].dir_option[ZCMD.arg2]->exit_info,
        #     EX_CLOSED);
        #   REMOVE_BIT(world[ZCMD.arg1].dir_option[ZCMD.arg2]->exit_info,
        #        EX_LOCKED);
        #   break;
        # case 2:
        #   SET_BIT(world[ZCMD.arg1].dir_option[ZCMD.arg2]->exit_info,
        #     EX_LOCKED);
        #   SET_BIT(world[ZCMD.arg1].dir_option[ZCMD.arg2]->exit_info,
        #     EX_CLOSED);
        #   break;
        # }
     #      last_cmd = 1;
        return (False, None)

    def removeObject(self, world, zone, state, nr, cmd):
        # case 'R': /* rem obj from room */
        #   if ((obj = get_obj_in_list_num(ZCMD.arg2, world[ZCMD.arg1].contents)) != NULL)
        #     extract_obj(obj);
        #   last_cmd = 1;
        #   break;
        return (False, None)

    def giveObject(self, world, zone, state, nr, cmd):
     #    case 'G':         /* obj_to_char */
     #      if (!mob) {
        # ZONE_ERROR("attempt to give obj to non-existant mob, command disabled");
        # ZCMD.command = '*';
        # break;
     #      }
     #      if (obj_index[ZCMD.arg1].number < ZCMD.arg2) {
        # obj = read_object(ZCMD.arg1, REAL);
        # obj_to_char(obj, mob);
        # last_cmd = 1;

     #  // zone load item

     #      } else
        # last_cmd = 0;
     #      break;
        return (False, None)



# CB: Is open-sourcing a specific input format to the expert system actually a fail? ;)
# (and maybe it should go in another module, not __init__)
#

# todo: move into ph.lang.document

from xml.dom import minidom
import re

class FileSource(str):
    def parse(self):
        return minidom.parse(open(self))
    def getName(self):
        return self
class InternalSource(str):
    def parse(self):
        return minidom.parseString(self)
    def getName(self):
        try: return self.__name
        except AttributeError:
            from md5 import md5
            self.__name = name = 'internal$%s' % md5(self).hexdigest()

# Basic Parsing Tools.
def parseSoftInt(value):
    if value not in ('', None):
        return int(value)

def parseBoolean(value):
    if isinstance(value, str):
        value = value.lower()
        if value in ('true', 'on', 'yes', '1'):
            return True
        if value in ('false', 'off', 'no', '0'):
            return False
        if value.isdigit():
            return int(value)

    raise TypeError(type(value))

def parseSoftBoolean(value):
    return None if value == '' else parseBoolean(value)

DICE_PATTERN = re.compile(r'(?P<number>\\d+)d(?P<size>\\d+)(?:\\+(?P<addend>\\d+))?')
def parseDice(dice):
    match = DICE_PATTERN.match(dice)
    if match is None:
        raise SyntaxError(dice)

    match = match.groupdict()
    return (match['number'], match['size'], match['addend'])

def parseWeaponDice(dice):
    match = DICE_PATTERN.match(dice)
    if match is None:
        raise SyntaxError(dice)

    match = match.groupdict()
    return (match['number'], match['size'])

# todo: allow '-' in place of '_'
def transferAttributes(node, object, attributes):
    for (name, function) in attributes.items():
        value = node.getAttribute(name)
        if value:
            setattr(object, name, function(value))

def getVnum(node, name = 'vnum'):
    return int(node.getAttribute(name))
def getWholeText(node):
    data = []
    for child in node.childNodes:
        if child.nodeType == child.TEXT_NODE:
            data.append(child.data)

    return ''.join(data)

# Loading.
NAMESPACES = dict(stuph      = 'telnet://stuphmud.net/3',
                  olc        = 'telnet://stuphmud.net/world/3',
                  search     = 'telnet://stuphmud.net/world/zifnab/search',
                  genspecact = 'telnet://stuphmud.net/world/tarn/genspecact',
                  player     = 'telnet://stuphmud.net/player/3',
                  wizard     = 'telnet://stuphmud.net/wizard/3')

class Loader:
    def __init__(self, environment):
        self.environment = environment

    def loadDocument(self, source):
        doc = source.parse()
        stuph = doc.documentElement
        assert stuph.localName == 'stuph' and stuph.namespaceURI == NAMESPACES['stuph']

        doc.source = source
        self.environment.addDocument(doc)

        # todo: map namespace-prefixes on upfront element and check on those for faster recognition

        for node in stuph.childNodes:
            if node.nodeType == node.ELEMENT_NODE:
                # detect node purpose and dispatch
                ns = node.namespaceURI
                name = node.localName

                if ns == NAMESPACES['olc']:
                    if name == 'zone':
                        zone = self.loadOlcZone(node)
                        self.environment.addOlcZone(zone)

                    elif name == 'houses':
                        pass

                elif ns == NAMESPACES['wizard']:
                    if name == 'command':
                        self.loadWizardCommand(node)
                    elif name == 'module':
                        self.loadWizardModule(node)

                elif ns == NAMESPACES['player']:
                    if name == 'trigger':
                        self.loadPlayerTrigger(node)
                    elif name == 'alias':
                        self.loadPlayerAlias(node)
                    elif name == 'profile':
                        self.loadPlayerProfile(node)
                    elif name == 'spell':
                        self.loadPlayerSpell(node)
                    elif name == 'mail':
                        self.loadPlayerMail(node)

                elif ns == NAMESPACES['stuph']:
                    if name == 'player':
                        self.loadStuphPlayer(node)

                    # Atypical and old-gen:
                    elif name == 'mobile':
                        self.loadStuphMobile(node)
                    elif name == 'item':
                        self.loadStuphItem(node)

        return doc

    parseContinentName = str
    zoneAttributes = dict(name = str, age = int, lifespan = int,
                          continent = parseContinentName)

    for flag in ('draft', 'immortal', 'clan', 'remort'):
        zoneAttributes[flag] = parseBoolean

    def loadOlcZone(self, node):
        # Commands and prototype data.
        zone = self.environment.ZoneData(vnum = getVnum(node))
        transferAttributes(node, zone, self.zoneAttributes)

        for child in node.childNodes:
            if child.nodeType == child.ELEMENT_NODE:
                name = child.localName
                if  name == 'commands':
                    self.loadOlcZoneCommands(zone, child)
                elif name == 'room':
                    self.loadOlcRoom(zone, child)
                elif name == 'mobile':
                    self.loadOlcMobile(zone, child)
                elif name == 'item':
                    self.loadOlcItem(zone, child)
                elif name == 'shop':
                    self.loadOlcShop(zone, child)

        return zone

    def loadOlcZoneCommands(self, zone, node):
        # Any and all zone commands.
        position = node.getAttribute('position')
        if position.isdigit():
            position = int(int)
        elif position == '':
            position = 'end'
        else:
            assert position in ('front', 'end', 'overwrite')

        for child in node.childNodes:
            if child.nodeType == child.ELEMENT_NODE:
                if child.tagName == 'room':
                    commands = list(self.loadOlcRoomCommands(zone, child))
                    zone.addCommands(position, commands)

    EXIT_NAMES = ['north', 'south', 'east', 'west', 'up', 'down']

    def loadOlcRoomCommands(self, zone, node):
        # Room commands container.
        room = zone.ZCmd.RoomCommand(zone = zone, vnum = getVnum(node))
        for child in node.childNodes:
            if child.nodeType == child.ELEMENT_NODE:
                name = child.tagName
                if name == 'item':
                    vnum = getVnum(child)
                    maximum = parseSoftInt(child.getAttribute('maximum'))

                    item = zone.ZCmd.ItemCommand(vnum = vnum, maximum = maximum)
                    self.loadOlcItemCommands(room, item, child)

                    yield item

                elif name == 'mobile':
                    yield self.loadOlcMobileCommands(room, child)
                elif name == 'remove':
                    yield zone.ZCmd.RemoveItem(room = room, vnum = getVnum(child, 'item'))

                elif name in self.EXIT_NAMES:
                    yield room.DoorState(room = room, direction = name,
                                         closed = parseSoftBoolean(child.getAttribute('closed')),
                                         locked = parseSoftBoolean(child.getAttribute('locked')))

    def loadOlcMobileCommands(self, room, node):
        # Mobiles loaded in room -- recursive on mount.
        mobile = room.zone.ZCmd.MobileCommand(vnum = getVnum(node),
                                              maximum = parseSoftInt(node.getAttribute('maximum')),
                                              room = room)

        for child in node.childNodes:
            if child.nodeType == child.ELEMENT_NODE:
                name = child.tagName
                if name == 'item':
                    position = child.getAttribute('position')

                    vnum = getVnum(child)
                    maximum = parseSoftInt(child.getAttribute('maximum'))

                    if position:
                        item = mobile.EquipmentCommand(vnum = vnum, maximum = maximum,
                                                    wearer = mobile, position = position)

                        self.loadOlcItemCommands(room, item, child)
                        mobile.addEquipment(item)
                    else:
                        item = room.zone.ZCmd.ItemCommand(vnum = vnum, maximum = maximum,
                                                          carrier = mobile)

                        self.loadOlcItemCommands(room, item, child)
                        mobile.addInventory(item)

                elif name == 'mount':
                    vnum = getVnum(child)
                    maximum = parseSoftInt(child.getAttribute('maximum'))

                    mount = room.zone.ZCmd.MobileCommand(vnum = vnum, maximum = maximum,
                                                         rider = mobile)

                    self.loadOlcMobileCommands(doc, room, child)
                    mobile.setMount(mount)

                # XXX this doesn't really exist.
                ##    elif name == 'unmount':
                ##        vnum = getVnum(child)
                ##        mobile.unmount(vnum)

        room.addMobile(mobile)
        return mobile

    def loadOlcItemCommands(self, room, item, node):
        # Items within containers -- meant for recursion.
        for child in node.childNodes:
            if child.nodeType == child.ELEMENT_NODE and child.tagName == 'item':
                vnum = getVnum(child)
                maximum = parseSoftInt(child.getAttribute('maximum'))

                subitem = room.zone.ZCmd.ItemCommand(vnum = vnum, maximum = maximum,
                                                     container = item)

                self.loadOlcItemCommands(room, subitem, child)
                item.addItem(subitem)

    parseRoomSector = str
    roomAttributes = dict(name = str, sector = parseRoomSector)

    # not house, housecrashsave, atrium, *, or clanhouseentrance
    for flag in ('dark', 'deathtrap', 'nomob', 'indoors', 'peaceful',
                 'soundproof', 'notrack', 'nomagic', 'tunnel', 'private',
                 'godroom', 'olc', 'implementorOnly', 'arena', 'noteleport',
                 'noquit', 'regeneration', 'nomortalpc', 'deleted'):
        roomAttributes[flag] = parseBoolean

    def loadOlcRoom(self, zone, node):
        room = zone.RoomData(vnum = getVnum(node), description = [],
                             extra_descriptions = {}, exits = {})

        transferAttributes(node, room, self.roomAttributes)

        for child in node.childNodes:
            if child.nodeType == child.ELEMENT_NODE:
                name = child.localName
                ns = child.namespaceURI

                if ns == NAMESPACES['search']:
                    # Q: Should this exist in room zone commands too?
                    self.loadOlcRoomSearches(room, child)

                elif name == 'description':
                    keywords = child.getAttribute('keywords')
                    if keywords:
                        room.extra_descriptions[keywords] = getWholeText(child)
                    else:
                        room.description.append(getWholeText(child))

                elif name in self.EXIT_NAMES:
                    self.loadOlcRoomExit(room, name, child)

        zone.addRoom(room)

    def loadOlcRoomSearches(self, room, node):
        if node.localName == 'on':
            search = room.SearchData()
            transferAttributes(node, search, dict(command = str, keywords = str,
                                                  match = str, reset = str))

            for child in node.childNodes:
                if child.nodeType == child.ELEMENT_NODE:
                    name = child.localName
                    if name == 'requires':
                        search.requires = Object(item = getVnum(child, 'item'))

                    elif name == 'act':
                        # todo: search object message api
                        messages = search('message', {})
                        messages[child.getAttribute('to')] = getWholeText(child)

                    elif name in ('open', 'close', 'toggle'):
                        self.loadOlcDoorSearch(search, child)
                    elif name == 'create':
                        self.loadOlcSpawnSearch(search, child)

                    elif name == 'teleport':
                        self.loadOlcTeleportSearch(search, child)

            room.addSearch(search)

    def loadOlcDoorSearch(self, search, node):
        door = Object()
        search('exits', {})[node.getAttribute('exit')] = door

        transferAttributes(node, door, dict(lock = parseBoolean))
        for child in node.childNodes:
            if child.nodeType == child.ELEMENT_NODE:
                # todo: search object message api
                messages = search('message', {})
                messages[child.getAttribute('to')] = getWholeText(child)

    def loadOlcSpawnSearch(self, search, node):
        item = Object(vnum = getVnum(node, 'item'))
        search['creation'].append(item)

        for child in node.childNodes:
            if child.nodeType == child.ELEMENT_NODE:
                name = child.localName
                if name == 'respawn':
                    item.respawn = True
                    self.loadOlcRespawnSearch(search('message', {}), child)

                elif name == 'act':
                    # todo: search object message api
                    message = search('message', {})
                    message[child.getAttribute('to')] = getWholeText(child)

    def loadOlcRespawnSearch(self, messages, node):
        for child in node.childNodes:
            if child.nodeType == child.ELEMENT_NODE and child.localName == 'act':
                messages[child.getAttribute('to')] = getWholeText(child)

    def loadOlcTeleportSearch(self, search, node):
        teleport = Object(vnum = getVnum(node, 'room'))
        search.teleport = teleport

        for child in node.childNodes:
            if child.nodeType == child.ELEMENT_NODE:
                name = child.localName
                message = search('message', {})

                if name in ('departure', 'arrival'):
                    # todo: search object message api -- XXX breaking convention
                    self.loadOlcTeleportMessageSearch(message.setdefault(name, {}), child)

    def loadOlcTeleportMessageSearch(self, message, node):
        for child in node.childNodes:
            if child.nodeType == child.ELEMENT_NODE and child.localName == 'act':
                message[child.getAttribute('to')] = getWholeText(child)

    exitAttributes = dict(room = int, key = int, keywords = str)

    for flag in ('isdoor', 'closed', 'locked', 'pickproof', 'hidden', 'nomagic'):
        exitAttributes[flag] = parseBoolean

    def loadOlcRoomExit(self, room, name, node):
        exit = room.ExitData(direction = name, description = [])
        transferAttributes(node, exit, self.exitAttributes)
        for child in node.childNodes:
            if child.nodeType == child.ELEMENT_NODE:
                if child.localName == 'description':
                    exit.description.append(getWholeText(child))

        room.addExit(name, exit)

    mobileAttributes = dict(keywords = str, name = str, sex = str, level = int, alignment = int,
                            gold = int, position = str, movement = str, armorclass = int,
                            hitroll = int, barehand = parseDice, hitpoints = parseDice,
                            experience = int, attack = str, strength = str, intelligence = str,
                            wisdom = str, dexterity = str, constitution = str, charisma = str,
                            special = str)

    # todo: mobile flags
    for flag in ('sentinel', 'scavenger', 'nobackstab', 'stayinzone', 'wimpy', 'memory',
                 'helper', 'nocharm', 'nosummon', 'nosleep', 'nobash', 'noblind', 'hunter',
                 'invulnerable', 'nofire', 'nosubdue', 'mobviolence', 'mount', 'silent',
                 'ignoresummons', 'nocorpse', 'noupdate', 'deleted'):
        mobileAttributes[flag] = parseBoolean

    mobileAffectionAttributes = dict((name, parseBoolean) for name in \
                                     ('blind', 'invisible', 'detectalign', 'detectinvis',
                                      'detectmagic', 'senselife', 'waterwalk', 'sanctuary',
                                      'cursed', 'infravision', 'poisoned', 'protectionfromevil',
                                      'protectionfromgood', 'sleep', 'notrack', 'flying',
                                      'fireshield', 'sneaking', 'hiding', 'phaseblur', 'charmed',
                                      'etherealized', 'secondattack', 'thirdattack', 'fourthattack',
                                      'meditating', 'reflect', 'absorb', 'entangled'))

    def loadOlcMobile(self, zone, node):
        mobile = zone.MobileData(vnum = getVnum(node))
        transferAttributes(node, mobile, self.mobileAttributes)

        position = node.getAttribute('position')
        if position:
            position = position.split('/', 2)
            if len(position) == 2:
                (mobile.position, mobile.default_position) = position
            else:
                mobile.position = position

        aggressive = node.getAttribute('aggressive')
        if aggressive:
            mobile.aggressive = aggressive # all|(evil,neutral,good,npc)

        for child in node.nodeType == node.ELEMENT_NODE:
            ns = child.namespaceURI
            name = child.localName

            if ns == NAMESPACES['genspecact']:
                continue

            if name in ['room_description', 'detailed_description']:
                # todo: further format room_description
                setattr(mobile, name, getWholeText(child))

            elif name == 'affections':
                af = mobile.Affections()
                mobile.affections = af

                transferAttributes(child, af, self.mobileAffectionAttributes)

        zone.addMobile(zone, mobile)

    itemAttributes = dict(keywords = str, name = str, cost = int,
                          minlevel = int, weight = int, timer = int)

    for flag in ('glowing', 'humming', 'norent', 'nodonate', 'noinvis', 'invisible',
                 'magic', 'cursed', 'blessed', 'antigood', 'antievil', 'antineutral',
                 'twohanded', 'flaming', 'poison', 'nosell', 'nomobtake', 'quest',
                 'nodecay', 'nolocate', 'noauction', 'shining', 'hidden', 'nodisarm',
                 'transparent', 'biodegradable', 'deleted', 'noidentify', 'itemnodecay'):
        itemAttributes[flag] = parseBoolean

    itemAnticlassAttributes = dict((name, parseBoolean) for name in \
                                   ('wizard', 'cleric', 'assassin', 'knight', 'ranger',
                                    'paladin', 'monk', 'ninja', 'druid'))

    itemWearableAttributes = dict((name, parseBoolean) for name in \
                                  ('take', 'finger', 'neck', 'body', 'head', 'legs',
                                   'feet', 'hands', 'arms', 'shield', 'about', 'waist',
                                   'wrist', 'wield', 'hold', 'ear', 'face', 'restraint'))


    ScrollAttrs = dict(spell_level = int,
                       spell1 = str,
                       spell2 = str,
                       spell3 = str)

    WandAttrs = dict(spell_level = int,
                     capacity = int,
                     charges = int,
                     spell = str)

    LiqContAttrs = dict(capacity = int,
                        contains = int,
                        liquid = str,
                        poisoned = parseBoolean)

    ITEM_TYPE_ATTRIBUTES = dict(light = dict(hours = int),
                                scroll = ScrollAttrs,
                                potion = ScrollAttrs,
                                pill = ScrollAttrs,
                                wand = WandAttrs,
                                staff = WandAttrs,
                                weapon = dict(spell = str,
                                              damage = parseWeaponDice,
                                              weapon = str),

                                armor = dict(armorclass = int),
                                trap = dict(biodegradable = parseBoolean),
                                container = dict(capacity = int,
                                                 key = int,
                                                 closeable = parseBoolean,
                                                 pickproof = parseBoolean,
                                                 closed = parseBoolean,
                                                 locked = parseBoolean,
                                                 container_type = str),

                                fountain = LiqContAttrs,
                                food = dict(satiation = int,
                                            poisoned = parseBoolean),

                                money = dict(worth = int),
                                portal = dict(min_level = int,
                                              max_level = int,
                                              room = int))

    ITEM_TYPE_ATTRIBUTES['liquid-container'] = LiqContAttrs

    def loadOlcItem(self, zone, node):
        item = zone.ItemData(vnum = getVnum(node), extra_descriptions = [])
        transferAttributes(node, item, self.itemAttributes)

        for child in node.nodeType == node.ELEMENT_NODE:
            name = child.localName

            # merge all of these descriptiony things
            if name in ['room_description', 'detailed_description']:
                # todo: further format room_description
                setattr(item, name, getWholeText(child))

            elif name == 'description':
                keywords = child.getAttribute('keywords')
                if keywords:
                    item.extra_descriptions[keywords] = getWholeText(child)

            elif name == 'anticlass':
                af = item.Anticlass()
                item.affections = af

                transferAttributes(child, af, self.itemAnticlassAttributes)

            elif name == 'wearable':
                bitv = item.Wearable()
                item.wearable = bitv

                transferAttributes(child, bitv, self.itemWearableAttributes)

            elif name == 'affects':
                self.loadOlcItemAffects(item, child)

            elif name == 'trap':
                trapType = child.getAttribute('type')
                if trapType:
                    item.trapType = trapType

        itemType = node.getAttribute('type')
        if itemType:
            item.type = itemType
            attributes = self.ITEM_TYPE_ATTRIBUTES.get(itemType)

            if attributes is not None:
                transferAttributes(node, item, attributes)

        zone.addItem(zone, item)

    def loadOlcItemAffects(self, item, node):
        for child in node.childNodes:
            if child.nodeType == child.ELEMENT_NODE and child.localName == 'apply':
                to = child.getAttribute('to')
                modifier = child.getAttribute('modifier')

                item.addAffect(to, modifier)

    def loadOlcShop(self, zone, node):
        pass

    def loadWizardCommand(self, node):
        # Todo: this node should go under a wizardly-module so that
        # it maybe compile-and-bind to the module's globals.
        action = node.getAttribute('action')
        verb = node.getAttribute('verb')
        level = node.getAttribute('level')

        for child in node.childNodes:
            if child.nodeType == child.ELEMENT_NODE and child.localName == 'code':
                source = child.getAttribute('source')
                codeType = child.getAttribute('type')
                code = getWholeText(child)

                cmd = self.environment.Wizard.Command((action, verb),
                                                      dict(level = level),
                                                      (source, codeType, code))
                self.environment.addModule(cmd)

                # Only process the first.
                break

    def loadWizardModule(self, node):
        pass

    def loadPlayerTrigger(self, node):
        pass
    def loadPlayerAlias(self, node):
        pass
    def loadPlayerProfile(self, node):
        pass
    def loadPlayerSpell(self, node):
        pass

    def loadStuphPlayer(self, node):
        pass
    def loadStuphMobile(self, node):
        pass
    def loadStuphItem(self, node):
        pass

# Environmental Model.
class Object:
    def __init__(self, **kwd):
        self.__dict__.update(kwd)
    def __call__(self, name, value):
        return self.__dict__.setdefault(name, value)
    def __getitem__(self, name):
        return self(name, [])

    Absent = object()
    def __repr__(self):
        def getValue(name):
            if callable(name):
                return name(self)

            value = self
            parts = name.split('.')
            first = parts[0]

            for name in parts:
                call = False
                if name.endswith('()'):
                    name = name[:-2]
                    call = True

                value = getattr(value, name, self.Absent)
                if value is self.Absent:
                    break

                if call:
                    value = value()

            return (first, value)

        def formatValue(v):
            # ???
            if isinstance(v, str):
                return str(v)
            return v

        attrs = getattr(self, '__attributes__', [])
        attrs = (getValue(n) for n in attrs)
        attrs = ' '.join('%s=%r' % (n, formatValue(v)) for (n, v) \
                         in attrs if v is not self.Absent)

        return '<%s%s%s>' % (self.__class__.__name__,
                             attrs and ' ' or '', attrs)

class Environment(Object):
    class ZoneData(Object):
        __module__ = __name__ + '.Environment'
        __attributes__ = ['vnum', 'name']

        class ZCmd:
            class RoomCommand(Object):
                __module__ = __name__ + '.Environment.ZoneData.ZCmd'
                __attributes__ = ['vnum']

                class DoorState(Object):
                    __module__ = __name__ + '.Environment.ZoneData.ZCmd.RoomCommand'
                    __attributes__ = ['direction', 'room']

                def addMobile(self, mobile):
                    self['mobiles'].append(mobile)
                def addItem(self, item):
                    self['items'].append(item)

            class MobileCommand(Object):
                __module__ = __name__ + '.Environment.ZoneData.ZCmd'
                __attributes__ = ['vnum']

                class EquipmentCommand(Object):
                    __module__ = __name__ + '.Environment.ZoneData.ZCmd.MobileCommand'
                    __attributes__ = ['vnum', 'position']

                    def addItem(self, item):
                        self['contents'].append(item)

                def addEquipment(self, item):
                    self['equipment'].append(item)
                def addInventory(self, item):
                    self['inventory'].append(item)

                def setMount(self, mount):
                    self.mount = mount

                ##    def unmount(self, vnum):
                ##        assert not hasattr(self, 'mount')
                ##        del self.mount

            class ItemCommand(Object):
                __module__ = __name__ + '.Environment.ZoneData.ZCmd'
                __attributes__ = ['vnum']

                def addItem(self, item):
                    self['contents'].append(item)

            class Position(Object):
                __module__ = __name__ + '.Environment.ZoneData.ZCmd'
                __attributes__ = ['position', lambda p: ('nr(commands)', len(p.commands))]

            class RemoveItem(Object):
                __module__ = __name__ + '.Environment.ZoneData.ZCmd'
                __attributes__ = ['vnum']

        class RoomData(Object):
            __module__ = __name__ + '.Environment.ZoneData'
            __attributes__ = ['vnum']

            class ExitData(Object):
                __module__ = __name__ + '.Environment.ZoneData.RoomData'
                __attributes__ = ['direction', 'room', 'key']

            class SearchData(Object):
                __module__ = __name__ + '.Environment.ZoneData.RoomData'
                __attributes__ = ['keywords', 'reset', 'command']

            def addExit(self, name, exit):
                self('exits', {})[name] = exit
            def addSearch(self, search):
                self['searches'].append(search)

        class MobileData(Object):
            __module__ = __name__ + '.Environment.ZoneData'

            class Affections(Object):
                __module__ = __name__ + '.Environment.MobileData'

        class ItemData(Object):
            __module__ = __name__ + '.Environment.ZoneData'

            class Anticlass(Object):
                __module__ = __name__ + '.Environment.ZoneData.ItemData'
            class Wearable(Object):
                __module__ = __name__ + '.Environment.ZoneData.ItemData'
            class Affection(Object):
                __module__ = __name__ + '.Environment.ZoneData.ItemData'

            def addAffect(self, to, modifier):
                self['affects'].append(self.Affection(to = to, modifier = modifier))

        def addCommands(self, position, commands):
            self['commands'].append(self.ZCmd.Position(position = position,
                                                       commands = commands))
        def addRoom(self, room):
            self['rooms'].append(room)
        def addMobile(self, mobile):
            self['mobiles'].append(mobile)
        def addItem(self, item):
            self['items'].append(item)

    def addDocument(self, doc):
        self('documents', {})[doc.source.getName()] = doc
        doc.environment = self

    def addOlcZone(self, zone):
        self('olc', Object(__attributes__ = ['zones'])) \
                        ('zones', {})[zone.vnum] = zone

    class Wizard:
        __module__ = __name__ + '.Environment'

        class Command(Object):
            __module__ = __name__ + '.Environment.Wizard'
            __attributes__ = ['verb', 'codeType', 'action']

            class Verb(Object):
                __module__ = __name__ + '.Environment.Wizard.Command'
                __attributes__ = ['verb']
            class Code(Object):
                __module__ = __name__ + '.Environment.Wizard.Command'
                __attributes__ = ['type', 'source']

            def __init__(self, xxx_todo_changeme, requirements, xxx_todo_changeme1):
                (action, verb) = xxx_todo_changeme
                (source, codeType, code) = xxx_todo_changeme1
                Object.__init__(self, verb = self.Verb(action = action, verb = verb),
                                requirements = Object(**requirements),
                                code = self.Code(type = codeType, source = source, code = code))

    def addModule(self, module):
        self['modules'].append(module)

# Compatability Mode (flatten for strict pickle):
##    Environment:
##        ZoneData:
##            ZCmd:
##                RoomCommand:
##                    DoorState
##                MobileCommand:
##                    EquipmentCommand
##                ItemCommand
##                Position
##                RemoveItem
##            RoomData:
##                ExitData
##                SearchData
##            MobileData:
##                Affections
##            ItemData:
##                Anticlass
##                Wearable
##                Affection
##
##        Wizard:
##            Command:
##                Verb
##                Code

ZoneData = Environment.ZoneData
ZCmd = ZoneData.ZCmd
RoomCommand = ZCmd.RoomCommand
DoorState = RoomCommand.DoorState
MobileCommand = ZCmd.MobileCommand
EquipmentCommand = MobileCommand.EquipmentCommand
ItemCommand = ZCmd.ItemCommand
Position = ZCmd.Position
RemoveItem = ZCmd.RemoveItem
RoomData = ZoneData.RoomData
ExitData = RoomData.ExitData
SearchData = RoomData.SearchData
MobileData = ZoneData.MobileData
Affections = MobileData.Affections
ItemData = ZoneData.ItemData
Anticlass = ItemData.Anticlass
Wearable = ItemData.Wearable
Affection = ItemData.Affection
Wizard = Environment.Wizard
Command = Wizard.Command
Verb = Command.Verb
Code = Command.Code

for _obj in (ZoneData, ZCmd, RoomCommand, DoorState, MobileCommand,
             EquipmentCommand, ItemCommand, Position, RemoveItem,
             RoomData, ExitData, SearchData, MobileData, Affections,
             ItemData, Anticlass, Wearable, Affection, Wizard, Command,
             Verb, Code):
    _obj.__module__ = __name__
    _obj.__name__ = _obj.__name__.split('.')[-1]

# Testing Front End.
from pickle import Unpickler, Pickler
import sys

class ClassfulUnpickler(Unpickler):
    def find_class(self, module, name):
        # This handles nested classes.
        module = module.split('.')
        __import__(module[0])
        mod = sys.modules[module[0]]

        for n in module[1:]:
            mod = getattr(mod, n)

        klass = getattr(mod, name)
        return klass

def main(argv = None):
    from optparse import OptionParser
    parser = OptionParser()
    parser.add_option('-p', '--pickle')
    parser.add_option('-u', '--unpickle')

    (options, args) = parser.parse_args(argv)

    global env
    if options.unpickle:
        env = ClassfulUnpickler(open(options.unpickle)).load()
    else:
        env = Environment()
        loader = Loader(env)

        for name in args:
            loader.loadDocument(FileSource(name))

        if options.pickle:
            Pickler(open(options.pickle, 'w')).dump(env)

if __name__ == '__main__':
    main()



from .dblib import *

defaultContinent = 0
validCmds = 'MOEPDRG'

class ZoneReader(RecordReader):
    # A ZoneReader always produces one record, so the record-set only
    # ever has one zone in it.

    VirtualNRName = 'vnum'
    nameOfRecordSet = nameOfCurrentRecord = 'zone'

    def startOfZone(self, vnum):
        self.zone = {'vnum': vnum}

    newRecord = startOfRecords = startOfZone

    def endOfZone(self):
        # Todo: Ensure presence of all attributes.
        pass

    endOfRecords = endOfZone

    def zoneField(self, name, value):
        self.zone[name] = value

    recordField = zoneField

    def zoneCommand(self, type, ifFlag, args):
        cmd = (type, ifFlag, args)
        r = self.zone # self.getCurrentRecord()

        if 'commands' in r:
            r['commands'].append(cmd)
        else:
            r['commands'] = [cmd]

    def disabledZoneCommand(self, line):
        ## print 'Disabled: '+repr(line)
        pass

    def parseZone(self):
        line=self.readLine()
        if line[0] != '#':
            raise FormatError()
        self.startOfZone(int(line[1:]))

        self.zoneField('name', self.tildeString())

        parts=self.readLine().split()
        l=len(parts)
        if l is 5:
            top, flags, lifespan, resetMode, continent = parts
            continent=int(continent)
        elif l is 4:
            top, flags, lifespan, resetMode = parts
            continent=defaultContinent
        elif l is 3:
            top, lifespan, resetMode = parts
            flags=''
            continent=defaultContinent
        else:
            raise FormatError('Not right number of args in: '+repr(parts))

        self.zoneField('top', int(top))
        self.zoneField('flags', ASCIIFlagDecode(flags))
        self.zoneField('lifespan', int(lifespan))
        self.zoneField('reset-mode', int(resetMode))

        self.zoneField('continent', continent)

        self.parseCommands()
        return self # Method chaining.

    parseRecord = parseFile = parseZone

    def parseCommands(self):
        while 1:
            line=self.readLine()
            if line=='':
                return self.EOFNoDollar()

            c=line[0]
            d=line[1]

            if c=='*':
                self.disabledZoneCommand(line)
                continue

            if c=='$' or c=='S':
                if d!='\n':
                    raise FormatError('Expected newline immediately following %c. Got %c.'%(c,d))
                break

            if c in validCmds and d==' ':
                parts=line.split()

                l=len(parts)
                if c=='R' or c=='G':
                    if l<4:
                        raise FormatError('Expected 4 arguments to %c.  Got %d'%(c,l))
                elif l<5:
                    raise FormatError('Expected 5 arguments to %c.  Got %d'%(c,l))

                c=parts[0]
                ifFlag=int(parts[1])
                args=parts[2:] # this is kept as string arguments, though they should be integers
            else:
                raise FormatError('Expected %c in %s'%(c,validCmds))

            self.zoneCommand(c, ifFlag, args)

        self.endOfZone()

    def consumeZone(self):
        zone=self.zone

        del self.zone
        del self.file

        return zone

    consumeRecordSet = consumeZone


class SelectiveZoneReader(ZoneReader):
    def __init__(self, file, include):
        ZoneReader.__init__(self, file)

        self.include=include

    def zoneField(self, name, value):
        if name in self.include:
            self.zone[name]=value # ZoneReader.zoneField(self, name, value)

"""
#<Zone No>
<Name>~
<Top as Integer> <Flags as ASCII flags> <Lifespan> <Reset mode> [<Continent>]

    Commands are as follows:
*
MOEPD <If-flag> <arg1> <arg2> <arg3>
$ or S

"""

## SQLObject representation.
DEFAULT_ZONE_DBFL = 'lib/etc/zones.db'

def openZoneTables(dbpath):
    from .sql import openDatabaseTables
    return openDatabaseTables \
        (dbpath, 'StuphZoneRecord', 'StuphZoneCommand',
         schema_package = 'sql.zone')

def convert_zone_row(zone, columns):
    'Convert Zone Record into a mapping suitable for SQLObject row representation.'
    row = {}

    for n in columns:
        if n in zone:
            row[n] = zone[n]

        # Naming exception.
        elif n in ('resetMode'):
            row['resetMode'] = zone['reset-mode']

    return row

def convert_command_row(command, columns):
    row = {}

    for n in columns:
        if n == 'type':
            row[n] = str(command[0])[0]
        elif n == 'ifFlag':
            row[n] = int(command[1])

        elif n == 'arg1':
            if len(command[2]) > 0:
                row[n] = int(command[2][0])
            else:
                row[n] = -1

        elif n == 'arg2':
            if len(command[2]) > 1:
                row[n] = int(command[2][1])
            else:
                row[n] = -1

        elif n == 'arg3':
            if len(command[2]) > 2:
                row[n] = int(command[2][2])
            else:
                row[n] = -1

    return row

def loadStuphZoneRecords(dbpath = DEFAULT_ZONE_DBFL, populate_with = None):
    'Open a database and return the relevent tables.  populate_with a file/name/sequence of .zon files.'
    from types import GeneratorType as generator

    StuphZoneRecord, StuphZoneCommand = openZoneTables(dbpath)

    if populate_with:
        def iterate_population(p):
            if type(p) in (list, tuple, generator):
                for x in p:
                    for y in iterate_population(x):
                        yield y
            elif type(p) in (str, file):
                # p is a filename
                yield ZoneReader(p).parseZone().consumeZone()

            elif isinstance(p, ZoneReader):
                # What if parseZone has already been performed? (track state)
                yield p.parseZone().consumeZone()

        zone_columns    =  list(StuphZoneRecord .sqlmeta.columns.keys())
        command_columns = list(StuphZoneCommand .sqlmeta.columns.keys())

        for zone in iterate_population(populate_with):
            zone_record = StuphZoneRecord(**convert_zone_row(zone, zone_columns))

            for c in zone.get('commands', ()):
                StuphZoneCommand \
                    (stuphZoneRecord = zone_record,
                     **convert_command_row \
                        (c, command_columns))

    return (StuphZoneRecord, StuphZoneCommand)

# FORMAT = '%(name)-35.30s #%(vnum)5d'
FORMAT   = '#[%(vnum)5d] %(name)-20.20s (%(top)5d:%(resetMode)02d:%(continent)02d) [%(flags)010d] %(commands)-68.68s'
PAGER    = 'less'

def zonelist(Z, fmt = None, selection = None, **kwd):
    fmt       = fmt or FORMAT
    selection = selection or Z.select(**kwd)
    columns   = list(Z.sqlmeta.columns.keys())

    def rowdict(row):
        d = {}

        for c in columns:
            d[c] = getattr(row, c)

        return d

    from os import popen
    popen(PAGER, 'w').write('\n'.join(fmt % rowdict(row) for row in selection))

def test(*args):
    keyByFilename = ('-key-by-filename' in args)

    if keyByFilename:
        args.remove('-key-by-filename')

    zone_mapping = {}

    for filename in args:
        zone = ZoneReader(filename).parseZone().consumeZone()

        if keyByFilename:
            # In case multiple different files define same zone vnum.
            zone_mapping[filename] = zone
        else:
            zone_mapping[zone['vnum']] = zone

    return zone_mapping

def main(opts = []):
    if '-test' in opts:
        opts.remove('-test')

        global zone_mapping
        zone_mapping = test(*opts)
    else:
        n = len (opts)
        i = 0

        populate_with = []
        dbfile = DEFAULT_ZONE_DBFL

        while i < n:
            a = opts[i]

            if a in ('-new', '-new-db', '-newdb'):
                dbfile = new_database_file()
            elif a in ('-db', '-dbfile', '-dbpath', '-db-file', '-db-path'):
                i += 1
                dbfile = opts[i]
            else:
                populate_with.append(a)

            i += 1

        return loadStuphZoneRecords(dbfile, populate_with = populate_with)

if __name__ == '__main__':
    from sys import argv

    if '-debug' in argv:
        from pdb import runcall
        main_app = main

        def main(*args):
            return runcall(main_app, *args)

        argv.remove('-debug')

    Z, Commands = main(argv[1:])

    import new
    Z.zonelist = new.instancemethod(zonelist, Z)



def bootPartnerAspect():
    # Conditionally booted Pentacle partner application off-thread.
    # Note: This should work in environments even if server is under autorun shell.
    from pentacle.bus import PENTACLE_PARTNER_NAME_ENV_VAR
    from os import environ
    if environ.get(PENTACLE_PARTNER_NAME_ENV_VAR):
        from pentacle.bus.partners import PartneredApplication
        from pentacle.runtime import nth
        nth(PartneredApplication.Main)



# Auto-Quest Generic Senarios Facility & Special Procedure.
# Copyright 2010-2021 Clint Banis.  All rights reserved.
#
from os.path import dirname, join as joinpath
from traceback import print_exc as traceback
from pprint import pformat
from sys import exc_info
from errno import ENOENT
from time import time as now
from io import StringIO as NewBuffer
from random import randint, choice
import re
from string import Template
import imp

try:
    from yaml import dump as dumpYaml
    from yaml import load as loadYaml
except ImportError:
    dumpYaml = loadYaml = False

from stuphos.runtime import newComponent, LookupClassObject as LookupSenarioObject
from stuphos.runtime.facilities import Facility
#from stuphos.zones.specials import SpecialInstantiator
from stuphos.management.config import loadConfig
from stuphmud.server.player import HandleCommandError
from stuphmud.server.player.shell import PromptFor
from stuphos.runtime.registry import getObject, delObject
from stuphos import enqueueHeartbeatTask
from stuphos import invokeTimeoutForHeartbeat

from world import mobile as MobilePrototype
from world import item as ItemPrototype
from world import room as RealRoom

# todo: wagering

class Object(dict):
    def __init__(self, **kwd):
        self.__dict__ = self
        self.update(kwd)
        self.validate()

    def validate(self):
        # Load internally.
        pass

    def __getitem__(self, name):
        return self.get(name)

def evaluate(value, **values):
    # Privileged: Execute as code for value!
    # Todo: utilize mud.runtime.sandbox module!
    co = compile(value, '<value>', 'eval')
    return eval(co, values, values)

def getFileText(filename):
    try: return open(filename).read()
    except IOError as e:
        if e.errno != ENOENT:
            (etype, value, tb) = exc_info()
            raise etype(value).with_traceback(tb)

def dumpYamlWithIndent(object, indent = 0):
    if type(indent) is int:
        indent = indent * ' '

    return '\n'.join(indent + line for line in dumpYaml(object).split('\n'))

def chance(n):
    return randint(1, 100) < n

def reloadThisModule():
    return imp.reload(__import__(__name__, globals(), locals(), ['']))

# Facility.
def getThisDir():
    return dirname(__file__)
def getConfigFile():
    return joinpath(getThisDir(), 'quests.cfg')

class QuestControl(Facility):
    NAME = 'Quest::Control'

    ACTIVATIONS_OBJECT_NAME = '%s::Activations' % NAME
    REWARDS_OBJECT_NAME = '%s::Rewards' % NAME

    class Manager(Facility.Manager):
        MINIMUM_LEVEL = Facility.Manager.IMPLEMENTOR
        VERB_NAME = 'qc*ontrol'

        def do_destroy(self, peer, cmd, args):
            for name in args:
                if name == 'activations':
                    delObject(QuestControl.ACTIVATIONS_OBJECT_NAME)
                    print('Destroyed activations.', file=peer)

                elif name == 'rewards':
                    delObject(QuestControl.REWARDS_OBJECT_NAME)
                    print('Destroyed rewards.', file=peer)

                else:
                    print('Unknown destroy option: %r' % name, file=peer)

        def do_reload(self, peer, cmd, args):
            s = self.facility.get()
            if s is not None:
                for name in args:
                    if name == 'senarios':
                        s.loadSenarios()
                        print('Reloaded senarios.', file=peer)

                    elif name == 'config':
                        s.loadConfig()
                        print('Reloaded configuration.', file=peer)

                    elif name == 'help':
                        s.loadHelp()
                        print('Reloaded help.', file=peer)

                    elif name == 'behavior':
                        s.loadBehavior()
                        print('Reloaded behavior.', file=peer)

                    else:
                        print('Unknown reload option: %r' % name, file=peer)

    @classmethod
    def create(self):
        return newComponent(self)

    def __instance_init__(self):
        self.loadConfig()
        self.loadSenarios()
        self.loadHelp()
        self.loadBehavior()

    def loadConfig(self):
        self.config = loadConfig(getConfigFile())
        self.cfgsection = self.config.getSection('QuestControl', **{'config-dir': getThisDir()})

    def loadSenarios(self):
        self.senarios = list(loadSenarios(self.config))

    def loadHelp(self):
        helpfile = self.getConfig('help-screen')
        screen = getFileText(helpfile) if helpfile else None

        helpfile = self.getConfig('help-topics')
        topics = []

        if helpfile:
            for (keywords, minlevel, text) in parseHelpFile(helpfile):
                topics.append((keywords.split(), text))

        self.help = Object(screen = screen, topics = topics)

    def loadBehavior(self):
        actions = []
        conversation = []
        self.behavior = Object(actions = actions, conversation = conversation)

        behaviorfile = self.getConfig('behavior-file')
        if behaviorfile:
            document = getFileText(behaviorfile)
            if document:
                for section in loadYaml(document):
                    for (name, group) in section.items():
                        if name == 'actions':
                            actions.append(group)
                        elif name == 'conversation':
                            for element in group:
                                conversation.append(ConversationElement(element['match'],
                                                                        element['response']))

    def getStatus(self):
        def analyze():
            yield '&yActivations&N'
            yield '&r===========&N'

            for a in self.activations:
                yield dumpYamlWithIndent(a, indent = 4)

        return '\r\n'.join(analyze()) + '\r\n'

    def getConfig(self, name):
        return self.cfgsection[name]

    def getHelpText(self, topic = None):
        if not topic:
            return self.help.screen
        else:
            if isinstance(topic, str):
                topic = topic.upper().split()
            elif isinstance(topic, (list, tuple)):
                topic = [t.upper() for t in topic]
            else:
                raise ValueError(topic)

            # Search keywords:
            # todo: allow quoted word groups
            for (names, text) in self.help.topics:
                for word in topic:
                    if word not in names:
                        break
                else:
                    return text

    # World Component Routines:
    # Note: technically, as a component, this class should override its __call__
    # method and dispatch through that to the activations, handling any possible
    # future events.
    def onSlayMobile(self, ctlr, victim, killer):
        for a in self.activations:
            o = getattr(a, 'onSlayMobile', None)
            if callable(o):
                o(victim, killer)

    def onDealDamage(self, ctlr, attacker, victim, dam, wtype):
        for a in self.activations:
            o = getattr(a, 'onDealDamage', None)
            if callable(o):
                o(attacker, victim, dam, wtype)

    def onPurgeMobile(self, ctlr, victim, wizard):
        for a in self.activations:
            o = getattr(a, 'onPurgeMobile', None)
            if callable(o):
                o(victim, wizard)

    def onDeathTrap(self, ctlr, victim, cause):
        for a in self.activations:
            o = getattr(a, 'onDeathTrap', None)
            if callable(o):
                o(victim, cause)

    # Senarios.
    def selectAllSenarios(self):
        return Selection(self.senarios)
    def selectSenariosByClass(self, cls):
        return Selection(self.senarios).selectByClass(cls)
    def selectSenariosByName(self, name):
        return Selection(self.senarios).selectByName(name)
    def findSenariosElligibleFor(self, player):
        return Elligibility(self.senarios, player)

    # Activations.
    @property
    def activations(self):
        return getObject(self.ACTIVATIONS_OBJECT_NAME,
                         create = list)

    def addActivation(self, a):
        self.activations.append(a)
    def removeActivation(self, a):
        actvtns = self.activations
        if a in actvtns:
            actvtns.remove(a)

    def getActivationsFor(self, player):
        return [a for a in self.activations if a.player is player]

    # Rewards.
    @property
    def rewards(self):
        return getObject(self.REWARDS_OBJECT_NAME,
                         create = self.loadRewards)

    def loadRewards(self):
        return dict()

    def saveRewards(self):
        # todo: backup
        pass

    def postReward(self, player, kitty):
        self.rewards.setdefault(player.idnum, []).append(kitty)
        self.saveRewards()

    def getRewardsFor(self, player):
        return self.rewards.get(player.idnum, [])

    def removeReward(self, player, kitty):
        try: self.rewards[player.idnum].remove(kitty)
        except (KeyError, IndexError): pass
        self.saveRewards()

    # A little SpecProc behavior.
    def getRandomActionGroup(self):
        actions = self.behavior.actions
        if actions:
            return choice(actions)

        return ()

    def getResponseAction(self, commstr):
        for element in self.behavior.conversation:
            for response in element(commstr):
                yield response

# Install.
QuestControl.manage()
# QuestControl.get(create = True)

class ConversationElement:
    def __init__(self, pattern, response):
        self.pattern = re.compile(pattern)
        self.template = Template(response)

    def __call__(self, commstr):
        match = self.pattern.match(commstr)
        if match is not None:
            # XXX -- what if it's not a dict match?
            yield self.template.substitute(match.groupdict())

# Could very well rely on stuphlib for this:
def parseHelpFile(filename):
    try: fl = open(filename)
    except IOError as e:
        if e.errno is ENOENT:
            return

        (etype, value, tb) = exc_info()
        raise etype(value).with_traceback(tb)

    while True:
        line = fl.readline()
        if line == '':
            break

        strippedLine = line.strip()
        if strippedLine == '$':
            return # end of records/file

        if strippedLine != '#':
            # Format error
            return

        keywords = ' '.join(fl.readline().upper().split())
        minlevel = int(fl.readline())
        content = NewBuffer()

        while True:
            line = fl.readline()
            if line.strip() == '~':
                break

            content.write(line)
            content.write('\r\n')

        content = content.getvalue()
        yield (keywords, minlevel, content)

def pushFlagStatus(flags, name, object, newValue):
    stack = getattr(object, 'flagstack_%s' % name, [])
    stack.append(getattr(flags, name))
    setattr(flags, name, newValue)
def popFlagStatus(flags, name, object, defaultValue):
    try: stack = getattr(object, 'flagstack_%s' % name)
    except AttributeError: pass
    else:
        try: setattr(flags, name, stack.pop())
        except IndexError: setattr(flags, name, defaultValue)

def startQuestStatus(player, object):
    pushFlagStatus(player.preferences, 'Quest', object, True)
def endQuestStatus(player, object):
    popFlagStatus(player.preferences, 'Quest', object, False)

class Kitty(Object):
    # External reward storage container.
    def validate(self):
        self.items = list(self.get('items', []))
        self.gold = int(self.get('gold', 0))
        self.points = dict(self.get('points', Object(quest = 0,
                                                     platinum = 0,
                                                     gold = 0,
                                                     silver = 0,
                                                     merit = 0)))

    def addItem(self, item):
        self.items.append(item)
    def addGold(self, gold):
        self.gold += gold
    def addPoints(self, type, amount):
        self.points[type] = self.points.get(type, 0) + amount

    # Serialization: serialize only item numbers (or obj_file_elem)
    # Gold and points are primitive

# Senarios Model.
SENARIO_SUBSTITUTES = {'builtin': __name__}

def loadQuest(quest, **kwd):
    qtype = quest.get('quest-type')
    if qtype:
        # Resolve name and senario object.
        qtype = Template(qtype).substitute(SENARIO_SUBSTITUTES)

        try: senarioClass = LookupSenarioObject(qtype)
        except ValueError: pass

        # Load document tree into object node.
        else: return senarioClass(**dict(quest, **kwd))

def loadSenarios(cfg):
    # Experimental:
    try: from dom.XML import FormatDocumentSource, XML
    except ImportError: pass
    else:
        XMLDOM = cfg.getSection('XML.DOM', **{'config-dir': getThisDir()})
        for option in XMLDOM:
            if option.startswith('filepath'):
                doc = open(XMLDOM[option]).read()
                doc = FormatDocumentSource(doc)
                doc = XML(doc)

                # Just return raw document.
                yield doc

    try: from yaml import load as loadYaml
    except ImportError: pass
    else:
        YAML = cfg.getSection('YAML', **{'config-dir': getThisDir()})
        for option in YAML:
            if option.startswith('filepath'):
                for quest in loadYaml(open(YAML[option])):
                    q = loadQuest(quest)
                    if q is not None:
                        yield q

# Time-Limits.
def seconds(value):
    return value
def minutes(value):
    return seconds(value) * 60
def hours(value):
    return minutes(value) * 60
def days(value):
    return hours(value) * 24
def weeks(value):
    return days(value) * 7
def months(value):
    return weeks(value) * 4
def years(value):
    return days(value) * 365.25
def mudhours(value):
    return seconds(75)
def muddays(value):
    return mudhours(35)
def mudmonths(value):
    return mudmonths(0)
def mudyears(value):
    return mudmonths(35)

timeFunctions = dict(seconds = seconds, minutes = minutes, hours = hours, days = days,
                     weeks = weeks, months = months, years = years,
                     ticks = mudhours, mudhours = mudhours, muddays = muddays,
                     mudmonths = mudmonths, mudyears = mudyears)

def evaluateTimeLimit(activation, value):
    if isinstance(value, str):
        ns = timeFunctions.copy()
        ns['player'] = activation.player

        return evaluate(value, **ns)

    if type(value) is not int and value is not None:
        raise ValueError(value)

    return value

# Requirements.
class Requirement:
    def __init__(self, code):
        self.code = code
    def __call__(self, player):
        return bool(evaluate(self.code, player = player))

    class Or:
        def __init__(self, decisions):
            self.decisions = decisions
        def __call__(self, player):
            for d in self.decisions:
                if d(player):
                    return True

            return False

    class Comparison:
        class Op:
            def __init__(self, attribute):
                self.attribute = attribute

        def __init__(self, cmp, value):
            self.cmp = cmp
            self.value = value

        def __call__(self, player):
            return self.cmp(player, self.value)

        class Equal(Op):
            def __call__(self, player, value):
                return getattr(player, self.attribute) == value

        class GreaterThan(Op):
            def __call__(self, player, value):
                return getattr(player, self.attribute) > value

        class GreaterThanOrEqual(Op):
            def __call__(self, player, value):
                return getattr(player, self.attribute) >= value

        class LessThan(Op):
            def __call__(self, player, value):
                return getattr(player, self.attribute) < value

        class LessThanOrEqual(Op):
            def __call__(self, player, value):
                return getattr(player, self.attribute) <= value

POPULAR_COMPARISON = {'min-level': Requirement.Comparison.GreaterThanOrEqual('level'),
                      'remort': Requirement.Comparison.GreaterThanOrEqual('remort'),
                      'gold': Requirement.Comparison.GreaterThanOrEqual('gold_on_hand')}

def buildDecisions(structure):
    if isinstance(structure, (list, tuple)):
        for req in structure:
            if isinstance(req, dict):
                for (name, value) in req.items():
                    if name == 'either':
                        yield Requirement.Or(buildDecisions(value))
                    elif name in POPULAR_COMPARISON:
                        yield Requirement.Comparison(POPULAR_COMPARISON[name], value)
                    else:
                        yield Requirement(value)

class Senario(Object):
    # Activation.
    class Activation(Object):
        def validate(self):
            self.__activated = True

        def isActivated(self):
            try: return self.__activated
            except AttributeError:
                return False

        def deactivate(self, reason = None):
            if self.isActivated():
                qc = QuestControl.get()
                if qc is not None:
                    qc.removeActivation(self)
                    endQuestStatus(self.player, self)

                    if reason:
                        peer = self.player.peer
                        if peer is not None:
                            # todo: string interpolation of reason
                            print(reason, file=peer)

                self.__activated = False

        TIMELIMIT_EXPIRATION_MESSAGE = '&M*&N &wThe time limit for the quest has expired!&N &M*&N'

        def onExpiration(self):
            self.deactivate(reason = self.TIMELIMIT_EXPIRATION_MESSAGE)

        def percentComplete(self):
            raise NotImplementedError

    DEFAULT_BEGIN_MESSAGE = '&M*&N &wWelcome to the quest!&N &M*&N\r\n'
    DEFAULT_END_MESSAGE = '&M*&N &wThe %(name)s quest is complete!&N &M*&N\r\n'

    activations = []

    def activate(self, player, **values):
        # Set up internal Activation state.
        a = self.Activation(player = player, senario = self, **values)
        self.setdefault('activations', []).append(a)

        qc = QuestControl.get()
        if qc is not None: # XXX else, error
            qc.addActivation(a)

        startQuestStatus(player, self)
        peer = player.peer
        if peer is not None:
            msg = self.get('begin-message', self.DEFAULT_BEGIN_MESSAGE)
            if msg:
                peer.page_string(msg)

        # timeout = evaluateTimeLimit(a, self.get('time-limit'))
        timeout = None
        a.time_limit = timeout
        a.start_time = now()

        if timeout is not None:
            invokeTimeoutForHeartbeat(timeout, a.onExpiration)

        # Now, charge the player anything set forth in requirements...

        return a

    # Sorting, Naming & Reports
    def nameMatch(self, name):
        return self.get('name', '').lower().split() == name.lower().split()
    def classMatch(self, cls):
        return self.get('class', '').lower().split() == cls.lower().split()

    def isElligible(self, player):
        for check in buildDecisions(self.get('requirements')):
            if not check(player):
                return False

        return True

    def shortName(self):
        cls = self.get('class')
        name = self.get('name', '?')
        return name if cls is None else '%s (%s)' % (name, cls)

    def shortDescription(self):
        return '[%(quest-type)10.10s] %(class)10.10s %(name)20.20s %(min-level)5.5s/%(max-level)5.5s' % self
    def detailedDescription(self):
        d = self.get('description')
        if type(d) in (list, tuple): d = '\r\n'.join(d)
        return '%s\r\n%s' % (d, self) if d else str(self)

    def __str__(self):
        return dumpYamlWithIndent(self, indent = 4)


class SharedSenario:
    name = 'shared'

    def message(self, player, message):
        from email.message import Message
        if isinstance(message, Message):
            if message.get_content_type() == 'text/html':
                return self.htmlMessage(player, message.get_payload())
            else:
                assert message.get_content_type() == 'text/plain'
                message = message.get_payload()

        message = "%s quest says, '%s'" % (player.name, message)
        for a in self.activations:
            if a.player.peer:
                print(message, file=a.player.peer)

    @runtime.available(runtime.Web.Adapter.SessionManager)
    def htmlMessage(adapter, self, player, message):
        # message = scrubMessage(message)

        message = '''\
            <div class="posted-quest">
                <h4><a href="/poke/{player.name}">{player.name}</a></h4>
                <h5>{timestamp} on {quest.name}</h5>
                <table>
                <tr>
                  <td>
                    <img src="/avatar/{player.name}">
                  </td>
                  <td>
                    <pre class="message">
                    {message}
                    </pre>
                  </td>
                </tr>
                </table>
            </div>
            '''.format(player = player, quest = self,
                       timestamp = 'past', message = message)

        for a in self.activations:
            p = a.player.peer
            if p:
                p = adapter.findSessionByPeer(p)
                if p:
                    p.postMessages(['html-panel', message])


# from stuphos.runtime.architecture.api import writeprotected
# class ProtectedSenario(writeprotected, SharedSenario, Senario):
#     def isElligible(self, player):
#         # todo: type check
#         return Senario.isElligible(self, player._object)
#     def activate(self, player, values = dict()):
#         # todo: type check
#         return Senario.activate(self, player._object, **values)

from stuphos.runtime.architecture import wrappedObject
class ProtectedSenarioClass(SharedSenario, Senario):
    pass

ProtectedSenarioClass._definition = wrappedObject._deny \
    (ProtectedSenarioClass,
     'Activation',
     'DEFAULT_BEGIN_MESSAGE',
     'DEFAULT_END_MESSAGE',
     'activations')

def ProtectedSenario(**info):
    return wrappedObject(ProtectedSenarioClass._definition,
                         ProtectedSenarioClass(**info))


# todo: move these into the Senario class.
class Selection:
    def __init__(self, senarios):
        self.senarios = list(senarios)

    def __iter__(self):
        return iter(self.senarios)
    def __str__(self):
        return self.shortDescription()
    def __bool__(self):
        return len(self.senarios) > 0

    def subselect(self, senarios):
        return self.__class__(senarios)
    def selectByName(self, name):
        return self.subselect(s for s in self.senarios if s.nameMatch(name))
    def selectByClass(self, cls):
        return self.subselect(s for s in self.senarios if s.classMatch(cls))

    def shortDescription(self):
        return '\r\n'.join(s.shortDescription() for s in self) + '\r\n'
    def detailedDescription(self):
        return '\r\n'.join(s.detailedDescription() for s in self) + '\r\n'
    def numberedDescription(self):
        return '\r\n'.join('#%3d %s' % (nr, self.senarios[nr].shortDescription()) \
                           for nr in range(len(self.senarios)))

class Elligibility(Selection):
    def __init__(self, senarios, player):
        Selection.__init__(self, (s for s in senarios if s.isElligible(player)))
        self.player = player

    def subselect(self, senarios):
        return self.__class__(senarios, self.player)

PROMPT_YESNO = "Please type 'yes' or 'no': "

def joinSenario(player, senario):
    def confirmJoin(peer, line):
        line = line.strip().lower()
        if line == 'yes':
            # Let the prompt cycle lift the writing mask.
            enqueueHeartbeatTask(activateForPlayer, senario, player)

        elif line not in ['n', 'no']:
            return True # re-prompt

    PromptFor(player.peer, confirmJoin, write_mode = True,
              message = PROMPT_YESNO)

def selectSenario(player, senarios):
    def performSelect(peer, line):
        line = line.lower().strip()
        if line.isdigit():
            try: s = senarios[int(line)]
            except IndexError: pass
            else:
                # Now pass to join-senario confirmation.
                joinSenario(player, s)
                return

        if line == 'quit':
            return
        if line:
            print('Invalid selection: %r' % line, file=peer)

        return True # re-prompt

    PromptFor(player.peer, performSelect, write_mode = True,
              message = selection.numberedDescription() + \
              "Enter a number (or 'quit'): ")

def leaveSenarios(player, activations):
    def performLeave(peer, line):
        line = line.lower().strip()
        if line == 'yes':
            for a in activations:
                a.deactivate(reason = 'You have left the %s senario.' % a.senario.name)

        elif line not in ['n', 'no']:
            return True # re-prompt

    PromptFor(player.peer, performLeave, write_mode = True,
              message = PROMPT_YESNO)

def activateForPlayer(senario, player):
    # This routine is kind of pointless: it's a nice alert, but
    # by no means asserts that state is alright, unless activate
    # did in fact clean up before re-raising its error.
    try: senario.activate(player)
    except:
        traceback()
        peer = player.peer
        if peer is not None:
            print('There was an error with the quest:', file=peer)
            HandleCommandError(peer, full_traceback = False)

def redeemReward(player, kitty):
    # bank?
    # messages?
    player.gold_on_hand += kitty.gold
    for item in kitty.items:
        if type(item) is int:
            ItemPrototype(item).instantiate(player)
        elif isinstance(item, ItemPrototype):
            item.location = player

    player.qpoints += kitty.points.get('quest', 0)
    # assert player.qpoints >= 0

# Special Procedures.
def parseArgstr(argstr):
    args = argstr.split() if argstr is not None else ()
    return (args[0].lower(),) + tuple(args[1:]) if args else args

possessive_pronoun = {'Male'   : 'his',
                      'Female' : 'her',
                      'Neutral': 'its'}
informal_pronoun = {'Male'    : 'lad',
                    'Female'  : 'lass',
                    'Neutral' : 'friend'}

def questingClosed(player, me):
    me.sayTo(player, 'Questing services are currently closed.')
    # todo: replace this with a proper act()
    me('emote folds %s arms.' % possessive_pronoun[me.sex])

CHECK_OUT_SYNTAX = "Type 'check out all|<number>' or 'list rewards' to claim your prize."

def getCommFromToString(comm, targetWord):
    return comm.lstrip()[len(targetWord):]

CONVERSATION_DELAY = 2

def QuestMaster(player, me, cmd, argstr):
    args = parseArgstr(argstr)
    if cmd.reserved:
        # Do periodic behavior to give us personality, and hint as tutorial.
        if chance(10):
            qc = QuestControl.get()
            if qc is not None:
                for action in qc.getRandomActionGroup():
                    me(action)

    if cmd.name in ('to', '>'):
        if args and player.find(args[0]) is me:
            qc = QuestControl.get()
            if qc is not None:
                for response in qc.getResponseAction(getCommFromToString(argstr, args[0])):
                    invokeTimeoutForHeartbeat(CONVERSATION_DELAY, me.sayTo, player, response)

        return False

    if cmd.name == 'help':
        qc = QuestControl.get()
        if qc is None:
            questingClosed(player, me)
        else:
            peer = player.peer
            if peer is not None:
                text = qc.getHelpText(args)
                if text is not None:
                    peer.page_string(text)
                    return True

    if cmd.name == 'join':
        qc = QuestControl.get()
        if qc is None:
            questingClosed(player, me)
        else:
            if qc.getActivationsFor(player):
                me.tell('You may only join one quest at a time (currently)', player)
            else:
                named = ' '.join(args)
                if not named:
                    me.tell("Join what, me %s?  Type 'list' for the senarios." % \
                            informal_pronoun[player.sex], player)
                else:
                    senarios = qc.findSenariosElligibleFor(player)
                    found = list(senarios.selectByName(named))
                    if not found:
                        me.tell('There are no %r senarios.' % named, player)
                    elif len(found) > 1:
                        me.tell('These are %r senarios:' % named, player)
                        selectSenario(player, found)
                    else:
                        found = found[0]
                        me.tell('Are you sure you want to join %s?' % found.shortName(), player)
                        joinSenario(player, found)

        return True

    if cmd.name == 'leave':
        qc = QuestControl.get()
        if qc is None:
            questingClosed(player, me)
        else:
            named = ' '.join(args).lower()
            if not named:
                me.tell("Leave which senario?  Type 'list joined' for senarios you've joined.", player)
            else:
                actvtns = []
                for a in qc.getActivationsFor(player):
                    if a.senario.nameMatch(named):
                        actvtns.append(a)

                if not actvtns:
                    me.tell('Not joined %r senarios.' % named, player)
                else:
                    me.tell('Leave this senario?' if len(actvtns) == 1 else 'Leave all these senarios?', player)
                    leaveSenarios(player, actvtns)

            return True

    if cmd.name == 'list':
        qc = QuestControl.get()
        if qc is None:
            questingClosed(player, me)
        else:
            peer = player.peer
            if peer is not None:
                if args:
                    a = args[0].lower()
                    if a in ('joined', 'active'):
                        # Special case: actually, show activated senarios for player.
                        actvtns = qc.getActivationsFor(player)
                        if not actvtns:
                            me.tell('You are not currently participating in any quests.', player)
                            me.tell("Type 'list' to see available senarios.", player)
                        else:
                            peer.page_string('\r\n'.join(a.senario.shortDescription() for a in actvtns) + '\r\n')

                        return True

                    elif a == 'rewards':
                        rewards = qc.getRewardsFor(player)
                        if not rewards:
                            me.tell('You have no pending rewards.', player)
                        else:
                            me.tell("These are your pending rewards.  Type 'check <reward>' to claim.", player)
                            peer.page_string('\r\n'.join('%3d %s' % (nr, rewards[nr]) \
                                                         for nr in range(len(rewards))) + '\r\n')

                        return True

                    elif a == 'all':
                        senarios = qc.selectAllSenarios()
                    elif a.startswith('class:'):
                        senarios = qc.selectSenariosByClass(a[6:])
                    else:
                        senarios = qc.selectSenariosByName(a)

                else:
                    senarios = qc.findSenariosElligibleFor(player)

                if not senarios:
                    me.tell('There are no available senarios.', player)
                else:
                    me.tell('These are the available senarios:', player)
                    peer.page_string(senarios.shortDescription())

        return True

    if cmd.name == 'tell':
        if args[:2] != ('me', 'about'):
            return False

        qc = QuestControl.get()
        if qc is None:
            questingClosed(player, me)
        else:
            peer = player.peer
            if peer is not None:
                named = ' '.join(args[2:])
                if not named:
                    me.tell('Tell you about which senario?', player)
                else:
                    senarios = qc.findSenariosElligibleFor(player).selectByName(named)
                    if not senarios:
                        me.tell('There are no available %r senarios!' % named, player)
                    else:
                        me.tell('These are the available %r senarios:' % named, player)
                        peer.page_string(senarios.detailedDescription())

        return True

    if cmd.name == 'check':
        if len(args) > 1 and args[0].lower() == 'out':
            if len(args) != 2:
                me.tell(CHECK_OUT_SYNTAX, player)
            else:
                qc = QuestControl.get()
                if qc is None:
                    questingClosed(player, me)
                else:
                    rewards = qc.getRewardsFor(player)
                    which = args[1]

                    if which == 'all':
                        for r in rewards:
                            redeemReward(player, r)
                            qc.removeReward(player, r)

                    elif not which.isdigit():
                        me.tell(CHECK_OUT_SYNTAX, player)
                    else:
                        which = int(which)
                        try: r = rewards[which]
                        except IndexError:
                            me.tell("Unknown reward %r.  Type 'list rewards'.", player)
                        else:
                            redeemReward(player, r)
                            qc.removeReward(player, r)

            return True

    if cmd.name == 'reload' and args and args[0] == 'quest-control':
        if player.level >= Facility.Manager.IMPLEMENTOR:
            qc = QuestControl.get()
            if qc is not None and qc.activations:
                # todo: don't allow operation to succeed?
                if player.peer is not None:
                    print('&RThere (were) %d quest-senarios activated!&N' % len(qc.activations), file=player.peer)

            QuestControl.destroy()
            try: me.special = reloadThisModule().QuestMaster
            except:
                if player.peer is not None:
                    HandleCommandError(player.peer)
            else:
                me.sayTo(player, 'RELOADED: %s' % QuestControl.get(create = True))

            return True

# todo: overload 'quest' command:
#   Allows cancelling of quest (if allowed)
#   Allows seeing what quests we're involved with



# Now a ph component that provides a runtime configuration.
#
# Separate from the framework/compartment core, used in an embedding context.
# Also separate from the (legacy) mud package that resides now in application.server.
#
# Copyright 2022 runphase.com .  All rights reserved.
# --

# MUD Runtime Core.
#    mud.runtime.core provides:
#       * access to the event bridge
#       * programming the virtual machine
#       * configuration access
#       * boot procedure
#
# This should probably be called 'bootstrap' (the runtime is the core)
from stuphos.runtime import BINDINGS, EVENT_NAMES, loadOnto, Binding
from stuphos.etc.tools.logs import *

from stuphos.management.config import PackageManager, getParentPath, joinpath
from stuphos.management.config import getStandardPath, loadConfig

from stuphos.runtime.registry import getObject as getRegistryObject
from stuphos.runtime.registry import delObject as deleteRegistryObject
from stuphos.runtime.registry import RegistryNotInstalled

from stuphos.runtime.plugins import installPlugins

from stuphos.etc.tools import isYesValue, isNoValue

__version__ = '0.9.1'


def getBridgeModule():
    ##    # First code to load the game module.
    ##    from pdb import run
    ##    run('from game import bridgeModule')

    try: from stuphos.system.api import game
    except ImportError:
        from stuphos.system import game

    return game.bridgeModule()

def callEventBridge(name, *args, **kwd):
    try: event = getattr(getBridgeModule(), name)
    except AttributeError: pass
    else:
        if callable(event):
            return event(*args, **kwd)

def getMudModule():
    from stuphmud import server as mud
    return mud

def getHeartbeat():
    return getBridgeModule().heartbeat

def enqueueHeartbeatTask(*args, **kwd):
    return getHeartbeat().enqueueHeartbeatTask(*args, **kwd)
def enqueueHeartbeatTaskDefault(*args, **kwd):
    return getHeartbeat().enqueueHeartbeatTaskDefault(*args, **kwd)
def deferredTask(*args, **kwd):
    # print(f'deferring task to heartbeat... {args}')
    return getHeartbeat().deferredTask(*args, **kwd)

enqueue = enqueueHeartbeatTask
executeInline = deferredTask

def inline(function):
    # Decorator
    def inlineWrapper(*args, **kwd):
        # print(f'inlining... {args}')
        return executeInline(function, *args, **kwd)

    try: inlineWrapper.__name__ = function.__name__
    except AttributeError: pass

    try: inlineWrapper.__doc__ = function.__doc__
    except AttributeError: pass

    return inlineWrapper

def invokeTimeoutForHeartbeat(timeout, function, *args, **kwd):
    return getHeartbeat().invokeTimeoutForHeartbeat(timeout, function, *args, **kwd)
def invokeTimeout(timeout, function, *args, **kwd):
    return getHeartbeat().invokeTimeout(timeout, function, *args, **kwd)

CONFIG_OBJECT_NAME = 'MUD::Configuration'
CONFIG_FILE = getStandardPath('etc', 'config.cfg')

def _createMUDConfig():
    from ...etc.tools import registerBuiltin
    o = loadConfig(CONFIG_FILE)
    registerBuiltin(o.addressor, 'configuration')
    registerBuiltin(o.addressorTruth, 'configurationTruth')
    registerBuiltin(runtime.MUD.Configuration, 'mudConfig')
    return o

def getConfigObject():
    # return mudConfig(loadConfig, CONFIG_FILE)
    return getRegistryObject(CONFIG_OBJECT_NAME,
                             create = _createMUDConfig)

def deleteConfig():
    # del runtime[mudConfig]
    return deleteRegistryObject(CONFIG_OBJECT_NAME)

def reloadConfigFile(filename, setOptions = False):
    global CONFIG_FILE
    CONFIG_FILE = filename
    deleteConfig()

    # debugOn()
    if setOptions is not False:
        # Load immediately and override with core cmdline opts.
        getConfigObject().loadSetOptionList(setOptions)


def getConfig(name, section = 'MUD'):
    try: return getConfigObject().get(name, section = section)
    except RegistryNotInstalled: pass
def getSection(section = 'MUD'):
    try: return getConfigObject().getSection(section)
    except RegistryNotInstalled: pass


# Boot Procedure.
SITE_PATH = getParentPath(__file__, 5) # XXX ??? (besides its configurable)
COMPONENTS_FILE = 'components.pth'

# EASY_SITE_PATH = joinpath(SITE_PATH, 'packages/third-party')
# EASY_INSTALL_FILE = 'easy-install.pth'

def getDefaultComponentsPath():
    return SITE_PATH, COMPONENTS_FILE

def installSite():
    # Manually search non-standard paths for .pth files.
    path = getConfig('components')
    if path is None:
        (path, file) = getDefaultComponentsPath()
    else:
        from os import sep
        i = path.rfind(sep)
        if i < 0:
            file = path
            path = '' # '.'
        else:
            file = path[i+1:]
            path = path[:i]

    PackageManager(path, file).install()
    # PackageManager(EASY_SITE_PATH, EASY_INSTALL_FILE).install()

def installBridge():
    bridgeModule = getBridgeModule()
    thisModule = getMudModule()

    # from mud.runtime import declare, DeclareEvent
    # declare(bridgeModule, EVENTS)
    loadOnto(BINDINGS, bridgeModule)

    thisModule.on = Binding(bridgeModule)
    thisModule.core = Binding(thisModule)

    ##    class bootStart(DeclareEvent):
    ##        Module = thisModule
    ##    class bootComplete(DeclareEvent):
    ##        Module = thisModule

    return bridgeModule

def installHost():
    from socket import error
    from errno import EADDRINUSE

    disabling = [EADDRINUSE]

    # error: [Errno 10013] An attempt was made to access a socket in a way forbidden by its access permissions
    try: from errno import WSAEACCES
    except ImportError: pass
    else: disabling.append(WSAEACCES)

    try:
        from stuphos.kernel import getHost
        from stuphos.kernel import NotConfigured
        try: getHost(create = True).start()
        except NotConfigured:
            logWarning('XMLRPC disabled.')

    except error as e:
        if e.args[0] not in disabling:
            from stuphos.etc.tools import reraiseSystemException
            reraiseSystemException()

        logWarning('Host port is in use -- XMLRPC disabled.  Please reconfigure!')

    except Exception:
        logException(traceback = True)


def installEnviron():
    # Configure the system/shell environment.
    from os import environ
    envCfg = getSection('Environment')
    for name in envCfg.options():
        environ[name] = envCfg.get(name)

def installSystemComponents():
    # Install system path components.
    systemComp = getSection('SystemComponents')
    systemPaths = []

    if systemComp is not None:
        for option in systemComp.options():
            if option == 'system-path' or \
               option.startswith('system-path.'):
                systemPaths.append(systemComp.get(option))

    from sys import path as syspathCore
    for path in systemPaths:
        if path not in syspathCore:
            syspathCore.append(path)

def installSystemPackages():
    systemPkgs = getSection('SystemPackages')
    packages = []

    if systemPkgs is not None:
        for option in systemPkgs.options():
            if option == 'package' or \
               option.startswith('package.'):
                pkg = systemPkgs.get(option)
                # todo: filter for duplicate packages?
                packages.append((option, pkg))

    for (n, pkg) in sorted(packages, key = lambda a_b: a_b[0]):
        try: __import__(pkg)
        except SyntaxError as e:
            print(f'[syntax error] {pkg}: {e}')

        except: logException(traceback = True)


def installJournal():
    try:
        from stuphos.management import syslog
        # debugOn()
        syslog.Journal.get(create = True)
    except:
        logException(traceback = True)


def installServices(thisModule):
    # Install VM
    # Install system paths
    # Install environ config
    # Install XMLRPC host service
    # Install journal
    # Import system modules (config)
    # Install Management modules
    # Install logging service

    # Create binding to bridge module.

    # Todo: actually make this a registered COM object!
    if not isYesValue(configuration.VM.disabled):
        try: from stuphos.kernel import Machine, Native
        except ImportError as e: # e.path
            logOperation(f'[services$install] {e}: {e.path}')
            logException(traceback = True)
        else:
            bridge = getBridgeModule()
            vm = bridge.heartbeat = Machine()

            runtime.System.Network.Core(lambda:thisModule)
            runtime.System.Bridge(lambda:bridge)
            runtime.System.Engine(lambda:bridge.heartbeat)

            thisModule.on.shutdownGame(vm.onShutdownGame)

            debug = configuration.AgentSystem.debug_natives

            if isYesValue(debug):
                Native._tracing = True
            elif debug is not None and isNoValue(debug):
                Native._tracing = False


    # runtime.components = registry.Access(runtime, 'registry')
    # runtime.builtin = builtin
    # builtin.runtime = runtime
    # runtime.system = builtin.system
    #     # runtime.system.core

    # runtime.core = runtime.system.module.stuphos ?
    #     runtime.core.mud = runtime.system.module.stuphos ?
    #     runtime.core.game = runtime.system.module.game
    #     runtime.core.world = runtime.system.module.world

    # from ph import psOpGameVsz
    # import builtins
    # builtins.psOpGameVsz = psOpGameVsz
    # import op
    # print(f'[vsz] pre-services {psOpGameVsz()}')

    # Pre-Management Set:
    installSystemComponents()
    installEnviron()
    installHost()

    # print(f'[vsz] post-host {psOpGameVsz()}') # +74Mb

    installJournal()

    # print(f'[vsz] post-journal {psOpGameVsz()}') # +74Mb

    # Note that this relies on core components working.
    # But the rest of the runtime should rely on system
    # packages, not the other way around.
    installSystemPackages()

    # print(f'[vsz] post-packages {psOpGameVsz()}')


    try:
        from stuphos.kernel import MachineCore
        MachineCore.systemCore()
    except:
        logException(traceback = True)


    from stuphos.etc.tools.logs import logger
    endpoint = getConfig('log-endpoint', 'MUD')
    receiver = logger._nullreceiver
    if endpoint:
        from xmlrpclib import ServerProxy
        receiver = ServerProxy(endpoint).logging.receive

    from ...etc.tools import registerBuiltin
    registerBuiltin(logger(receiver), 'log')


    try: from stuphmud.server.player.interfaces.code.girl import initCommands
    except ImportError: pass
    else: initCommands()


    try: from stuphos.management import initForCore
    except ImportError: pass
    else:
        # Hack -- fixup core plugins before managed components.
        try:
            import stuphmud.server.player.db.managedfs
            from stuphmud.server.player.commands.gc import wizard
        except:
            # XXX :skip: Why doesn't this print the exception causes?
            # import traceback
            # traceback.print_exc()

            logException(traceback = True)

        initForCore()


    # Todo: integrate this into management initForCore?
    # Note: AgentSystem (plugin) relies on webserver, which is
    # a management component.  Use Management:webserver-object.
    # debugOn()
    installPlugins()


def installWorld():
    try: from stuphmud.server.zones import initForCore
    except ImportError: pass
    else: initForCore()

    from stuphmud.server.player.commands import installCommands
    installCommands()


def installSystemConfig():
    from os import getenv
    config = getenv('STUPHOS_CONFIG_ENV')
    if config:
        # Load --set-option from command line for initial config.
        reloadConfigFile(config, setOptions = core.cmdln['options'].set_option)


# Event Bridge.
def bootStart(configFile, setOptions):
    # Note: this function must return the bridge, or
    # the rest of the extension is not installed.

    # todo: import more of management
    # import mud.tools.debug


    # Least configurable:
    # installSystemConfig()

    bridge = installBridge()


    from stuphos.runtime.registry import getRegistry
    thisModule = getMudModule()
    getRegistry(create = thisModule.on) # Binding passed for shutdown-game registration.

    if configFile:
        # Note: this is necessary in order to activate 'configuration' builtin.
        reloadConfigFile(configFile, setOptions = setOptions)


    installSite()


    import sys
    sys._ph_security_package = configuration.Security.ph_security_package


    # Start reliance on VM path packages and rest of site.
    try:
        installServices(thisModule)
        installWorld()

        from stuphos.management.reboot import StartRecovery
        StartRecovery()

    except:
        logException(traceback = True)
    finally:
        return bridge

def bootComplete(timing):
    (secs, usecs) = timing
    from stuphos.management.reboot import CompleteRecovery
    CompleteRecovery()



'High-level struct implementation of player file format.'
from hstruct import *

MAX_NAME_LENGTH  = 20
MAX_EMAIL_LENGTH = 80
MAX_PWD_LENGTH   = 10
MAX_AFFECT       = 32
MAX_COLOR_SET    = 32
HOST_LENGTH      = 40
MAX_SKILLS       = 200
MAX_TONGUE       = 3

class CharSpecialData(Struct):
	alignment     = Type.Int
	idnum         = Type.Long
	actflags      = Type.Long
	affected      = Type.Long
	saving_throws = Type.Short[5]

class PlayerSpecialData(Struct):
	skills          = Type.Byte[MAX_SKILLS+1]
	padding         = Type.Byte
	talks           = Type.Byte[MAX_TONGUE]
	wimp_level      = Type.Int
	freeze_level    = Type.Byte
	invis_level     = Type.Short
	load_room       = Type.Int
	preferences     = Type.Long
	preferences2    = Type.Long
	displayflags    = Type.Long
	bad_password    = Type.UnsignedByte
	conditions      = Type.Byte[3]

	spells_to_learn = Type.Int
	marital_status  = Type.Int
	num_marriages   = Type.Int
	married_to      = Type.Long
	spell_attack    = Type.Int[2]
	remort_points   = Type.Int

class AbilityData(Struct):
	strength           = Type.Byte
	strength_addition  = Type.Byte
	intelligence       = Type.Byte
	wisdom             = Type.Byte
	dexterity          = Type.Byte
	constitution       = Type.Byte
	charisma           = Type.Byte

class PointData(Struct):
	mana       = Type.Short
	max_mana   = Type.Short
	hit        = Type.Short
	max_hit    = Type.Short
	move       = Type.Short
	max_move   = Type.Short

	deaths     = Type.Short
	mkills     = Type.Short
	pkills     = Type.Short
	dts        = Type.Short
	qpoints    = Type.Short
	armor      = Type.Short

	gold       = Type.Int
	bank       = Type.Int
	experience = Type.Int

	hitroll    = Type.Short
	damroll    = Type.Short

class AffectedType(Struct):
	type      = Type.Short
	duration  = Type.Short
	modifier  = Type.Byte
	location  = Type.Byte
	bitvector = Type.Long
	next      = Type.Pointer

class PlayerData(Struct):
	name            = Type.String[MAX_NAME_LENGTH +1]
	email           = Type.String[MAX_EMAIL_LENGTH+1]
	remort          = Type.Int
	race            = Type.Int
	clan            = Type.Int
	clanrank        = Type.Int
	page_length     = Type.Int
	save_room       = Type.Int
	sex             = Type.Byte
	chclass         = Type.Byte
	level           = Type.Byte
	hometown        = Type.Short
	birth           = Type.Int
	played          = Type.Int
	weight          = Type.UnsignedByte
	height          = Type.UnsignedByte

	pwd             = Type.String[MAX_PWD_LENGTH+1]

	char_specials   = Type.Struct(CharSpecialData)
	player_specials = Type.Struct(PlayerSpecialData)
	abilities       = Type.Struct(AbilityData)
	points          = Type.Struct(PointData)
	affected        = Type.Struct(AffectedType)[MAX_AFFECT]
	color           = Type.Int[MAX_COLOR_SET]

	last_login      = Type.Int
	last_logon      = Type.Int
	host            = Type.String[HOST_LENGTH+1]



# Player Commands Package.
from stuphmud.server.player import ACMD, ACMDLN, Option, getSharedScope
from stuphos.etc.tools import logException
from .mime import loadMessageFile
from os import listdir
from os.path import join as joinpath
import string

def installCommands():
    from stuphos.management.services import getEnablingSection
    isEnabled = getEnablingSection('Interpreter')

    if isEnabled('rich-editor'):
        import stuphmud.server.player.editor
    if isEnabled('player-notebook'):
        import stuphmud.server.player.notebook

    if isEnabled('wizard-gc'):
        from stuphmud.server.player.commands.gc import wizard
    if isEnabled('checkpointing'):
        from stuphmud.server.player.commands.gc import system
        system.EnableCheckpointing()

    try: loadFilesystemCommands()
    except: logException(traceback = True) # Is this right here?

def loadFilesystemCommands():
    from stuphos import getSection
    commands = getSection('VerbCommands')

    # Define commands from files in folder.
    for option in commands.options():
        if option.startswith('commands-folder'):
            folder = commands.get(option)
            for filename in listdir(folder):
                if filename.endswith('.mime'):
                    filename = joinpath(folder, filename)
                    with open(filename) as message:
                        loadMessageFile(message)

# Todo: Runtime API.
class BuiltCommandError(SyntaxError):
    def __init__(self, syntax, declaration, tb):
        self.syntax = syntax
        self.declaration = declaration
        self.tb = tb

def programVerbCommand(verbCode, language, source, options = [],
                       minlevel = None, group = None,
                       shlex = None, decorators = []):

    assert language == 'python'

    # Format parameters and verb name.
    parameters = 'player, command'

    if '*' not in verbCode:
        verbCode += '*'

    # Make command registrator.
    register = ACMDLN(verbCode,
                      *(options or []),
                      **dict(shlex = shlex))

    # Decide function name holding command code.
    valid = string.lowercase + string.uppercase + string.digits

    def formatWord(w):
        # Strip invalid characters.
        w = ''.join(c for c in w if c in valid)
        return w[:1].upper() + w[1:].lower()

    functionName = verbCode.replace('_', '-').split('-')
    functionName = 'do%s' % ''.join(formatWord(w) for w in functionName)

    def buildCommand(functionName, parameters, statements):
        # Compile the command statements and return the function.
        statements = statements.replace('\r', '')
        declaration = 'def %s(%s):\n    %s\n' % \
                       (functionName, parameters,
                        '\n    '.join(statements.split('\n')))

        # Declare function symbol in locals, but bind it to a
        # shared global scope so it has access later.
        ns = dict()

        # scope = player.interpreter.scope.__dict__

        ##    from mud.player import getPlayerScope
        ##    scope = getPlayerScope(player).__dict__

        scope = getSharedScope().__dict__

        try: exec(declaration, scope, ns)
        except SyntaxError as syntax:
            from sys import exc_info
            (_, _, tb) = exc_info()
            raise BuiltCommandError(syntax, declaration, tb)

        return ns[functionName]

    # Perform build.
    function = buildCommand(functionName, parameters, source)

    # Wrap.
    for d in decorators:
        function = d(function)

    register(function)
    return function



# MUD Runtime Object Registry.
# Copy 2021 runphase.com .  All rights reserved.
#
# Todo: get this from common tools.
_undefined = object()

class RegistryNotInstalled(RuntimeError):
    pass
class ObjectAlreadyRegistered(RuntimeError):
    pass
class DoesNotExistError(KeyError):
    pass

class Access(object): # todo: inherit from runtime.structure.Object? and builtin.object
    # VHLL object access interface and runtime base class.
    doesNotExist = DoesNotExistError

    class Node(object):
        def __init__(self, parent, name, call):
            self._parent = parent
            self._name = name
            self._call = call

        def _buildName(self):
            if self._parent:
                return '%s::%s' % (self._parent._buildName(), self._name)

            return self._name

        def __str__(self):
            return self._buildName()
        __repr__ = __str__

        def __getattr__(self, name):
            if not name.startswith('_'):
                # The basic goal here is to return a sub-node with priority
                # over attributes on the real, registered object, but support
                # dynamically access to those members secondarily.  This means
                # dynamic access to those members, except where eclipsed by
                # an inferior sub-node.
                #
                # This works effectively if the sub-objects are initialized
                # before the main object.  Otherwise, you'll have to use the
                # explicit get/registerObject toplevel routines if you know
                # for sure that it has a similarly named attribute.
                sub = Access.Node(self, name, self._call) # for reference.

                # Try to get the inferior object as named.
                if getObject(sub._buildName()) is None:
                    # Try to find an attribute on this named object.
                    o = getObject(self._buildName())
                    if o is not None:
                        # Only if there is an explicit attribute, and this isn't
                        # eclipsed by a registered sub-node object.
                        try: return object.__getattribute__(o, name)
                        except AttributeError:
                            pass # in this case, use unregistered sub-node:

                # Return the inferior node (not the object).
                # This means explicitly registered:
                #   Object::API
                #
                # will override the runtime['Object'].API real attribute,
                # that would be dereferenced above.
                #
                # By returning the node, instead of the object, sub-inferior
                # objects can still be referenced, even if there's an object
                # at Object, Object::API and Object::API:SubApi.
                #
                # This way, the real registered object is never returned by
                # this attribute lookup, only ever nodes.  Use explicit
                # getObject dereferencing to achieve this:
                #
                #   runtime['Object'].API
                #
                return sub

            return object.__getattribute__(self, name)

        def __call__(self, *args, **kwd):
            # runtime.Resource.Pool.Master(dict)
            # Q: Should this use registerObject? (to raise an error..)
            return self._call(self, *args, **kwd)

        def __enter__(self):
            return self
        def __exit__(self, *args):
            pass

    def __getattr__(self, name):
        # Create Top-level Accessor.
        # XXX :skip: Also, prevents top-level object names from being accessed
        # directly, this way (because Access is not an Access.Node).
        #
        return object.__getattribute__(self, name) \
               if name.startswith('_') \
                  else Access.Node(None, name, self._callCreate)

    def __getitem__(self, node):
        if isinstance(node, self.Node):
            node = node._buildName()

        if not isinstance(node, str):
            raise ValueError(node)

        return getObject(node)

    def __delitem__(self, node):
        if isinstance(node, self.Node):
            node = node._buildName()

        if not isinstance(node, str):
            raise ValueError(node)

        return delObject(node)

    def _callCreate(self, node, create = None, *args, **kwd):
        return getObject(node._buildName(),
                         create = lambda:create(*args, **kwd))

    def _doesNotExistError(self, name):
        raise DoesNotExistError(name)

    def _callNoCreate(self, node):
        return self._callCreate(node, self._doesNotExistError, node._buildName())

    def _callCall(self, node, *args, **kwd):
        # Call the registry object as a method with arguments.
        # Object must exist in registry or an error is raised.
        method = self._callNoCreate(node)
        return method(*args, **kwd)

    def _callCallIf(self, node, *args, **kwd):
        # Call the registry object as a method with arguments.
        # Nothing happens if object doesn't exist.
        try: method = self._callNoCreate(node)
        except DoesNotExistError: pass
        else: return method(*args, **kwd)

    def _callOn(self, node, client, *args, **kwd):
        # Wrap the client method with the callable registry object constructor.
        # Object must exist in registry or an error is raised.
        #
        # Note: this re-interprets the access node as the parent being an object
        # that has an attribute given by the terminal node name that is a Binding,
        # and then dereferences the binding's decorator (and calls it).

        parent = self._callNoCreate(node._parent)
        method = getattr(parent, node._name).on
        return method(client, *args, **kwd)

    def api(self, fon):
        return registerApi(fon)

    class _method(object):
        def __init__(self, access, method):
            self._access = access
            self._method = method

        def __getattr__(self, name):
            return self._access.Node(None, name, self._method)

    @property
    def call(self):
        return self._method(self, self._callCall)
    @property
    def callIf(self):
        return self._method(self, self._callCallIf)
    @property
    def on(self):
        return self._method(self, self._callOn)

    def __call__(self, node, *args, **kwd):
        def call(function):
            default = kwd.pop('default')
            object = self[node]
            if object is None:
                return default

            return function(object, *args, **kwd)

        return call

    def available(self, node, default = None):
        # todo: make this an accessor?
        def make(function):
            # XXX :skip: instance argument positioning.
            # todo: bind an availability decoration before object registration?
            def call(*args, **kwd):
                kwd = dict(default = default, **kwd)
                result = self(node, *args, **kwd)
                return result(function)
            return call
        return make


    def requireCall(self, node, *args, **kwd):
        def call(function):
            object = self[node]
            if object is None:
                raise AssertionError(f'Not available: {node}')

            return function(object, *args, **kwd)

        return call

    def require(self, node):
        def make(function):
            def call(*args, **kwd):
                result = self.requireCall(node, *args, **kwd)
                return result(function)
            return call
        return make


def lookup(accessor):
    '''
    @lookup(runtime)
    def dereference(node, *args, **kwd)
        return lambda *secondArgs, **secondKwd: \
            runtime[node](*args, **kwd)

    '''

    def bindToMethod(method):
        return accessor._method(accessor, method)

    return bindToMethod


# Construction:
def _createRegistry():
    return dict()

# The registry must be created initially.
# ---
# The registry storage object is a part of the runtime object,
# which is a component of the system module:
#    (systemModule->)runtime->(registryAccess)
#
# Both of these objects are composed with the following routines.

RUNTIME_OBJECT_NAME = 'runtime'
def getRuntimeObject():
    import sys as systemModule

    # Try to get from container module cache.
    try: return getattr(systemModule, RUNTIME_OBJECT_NAME)
    except AttributeError:
        pass

    # The Runtime class implementation is in a superior module,
    # (which derives from registry.Access)
    from . import MudRuntime
    runtime = MudRuntime()
    runtime.system = runtime.System #: Cache this node.

    # Put into runtime container module.
    setattr(systemModule, RUNTIME_OBJECT_NAME, runtime)

    from stuphos.etc.tools import registerBuiltin
    registerBuiltin(runtime, RUNTIME_OBJECT_NAME)
    registerBuiltin(runtime, 'components')

    return runtime

# XXX :skip: Want to set 'registry' on runtime, but the attribute lookup dynamically
# builds an Access.Node for anything that doesn't start with '_'
RUNTIME_REGISTRY_NAME = '_registry'
def getRegistry(create = False):
    runtimeObject = getRuntimeObject()

    try: registry = getattr(runtimeObject, RUNTIME_REGISTRY_NAME)
    except AttributeError:
        if create not in (None, False, 0):
            registry = _createRegistry()
            setattr(runtimeObject, RUNTIME_REGISTRY_NAME, registry)
            # runtimeObject.registry = registry

            # Hook shutdown game event with registry-cleanup function:
            #   This expects a certain kind of object (Binding) to
            #   be passed as 'create' parameter.  Otherwise, ignores.
            #
            #   (Admittedly, a weird convention).
            #
            try: create.shutdownGame(deleteAllObjects)
            except AttributeError:
                pass
        else:
            registry = None

    return registry

# Access API:
def getObject(name, create = None, **kwd):
    reg = getRegistry()
    if reg is None:
        raise RegistryNotInstalled('Registry not installed.')

    obj = reg.get(name, _undefined)
    if obj is _undefined:
        if callable(create):
            obj = reg[name] = create(**kwd)
        else:
            obj = None

    return obj

def setObject(name, value):
    reg = getRegistry()
    if reg is not None:
        if isinstance(name, Access.Node):
            name = name._buildName()

        reg[name] = value
    else:
        raise RegistryNotInstalled('Registry not installed.')

def registerObject(name, value):
    reg = getRegistry()
    if reg is not None:
        if isinstance(name, Access.Node):
            name = name._buildName()

        if name in reg:
            raise ObjectAlreadyRegistered(name)

        reg[name] = value
    else:
        raise RegistryNotInstalled('Registry not installed.')

def registerApi(fon, apiObject = None): # function-or-name
    if isinstance(fon, Access.Node):
        fon = fon._buildName()

    if isinstance(fon, str):
        if apiObject is not None:
            # Just directly register the object as an api.
            registerObject(fon, apiObject)
            return

        def makeNamedApiRegistry(object):
            registerObject(fon, object)
            return object

        return makeNamedApiRegistry

    # Expect a decorator style.
    assert apiObject is None

    def makeDefaultApiRegistry(object):
        # This style looks for a pre-programmed name.
        registerObject(object.NAME, object)
        return object

    return makeDefaultApiRegistry

def callObject(name, *args, **kwd):
    object = getObject(name)
    if callable(object):
        return object(*args, **kwd)

def callObjectMethod(name, methodName, *args, **kwd):
    object = getObject(name)
    if object is not None:
        try: method = getattr(object, methodName)
        except AttributeError: pass
        else: return method(*args, **kwd)

def _doDelete(object):
    # Todo: Maybe call this __unregister__?
    try: delete = getattr(object, '__registry_delete__', None)
    except ValueError:
        # This occurs if the object is a builtin entity handle
        # that has become invalid (through game-play).
        delete = None

    if not callable(delete):
        return True

    result = delete()
    if result is None:
        return True

    return bool(result)

def delObject(name, registry = None):
    if registry is None: # Err, why am I doing this?  Don't do this...
        registry = getRegistry()

    if registry is not None and name in registry:
        if _doDelete(registry[name]):
            del registry[name]
            return True

    return False

deleteObject = delObject

def deleteAllObjects():
    from traceback import print_exc as traceback

    reg = getRegistry()
    for obj in list(reg.values()):
        try: _doDelete(obj)
        except:
            traceback()

    reg.clear()

# Runtime-Access Usage:
##    # Creation:
##    r_pool = runtime.System.Resource.Pool(dict)
##
##    # Deletion:
##    del runtime['System::API']
##    del runtime[runtime.System.API]
##
##    # Method-decorator-based registration.
##    @registerApi('System::API')
##    class SystemAPI:
##        pass
##
##    @registerApi(runtime.System.API)
##    class SystemAPI:
##        pass
##
##    @registerApi
##    class SystemAPI:
##        NAME = 'System::API'
##
##    # Runtime-access-decorator-based registration.
##    @runtime.api('System::API')
##    class SystemAPI:
##        pass
##
##    @runtime.api(runtime.System.API)
##    class SystemAPI:
##        pass
##
##    @runtime.api
##    class SystemAPI:
##        NAME = 'System::API'
##
##    # Exact object lookup:
##    runtime[runtime.System.API]
##    runtime['System::API']

# Application example:
##    # Cache this lazy accessor for immediate invocation:
##    cmApi = runtime.system.CommandMessaging.API
##    def HandlePayload(self, payload):
##        cmApi.ExecuteTrustedCommandMessage(self.getUrlBase(), payload)

# Todo: Allow another kind of access:
# '@penobscotrobotics.us/modules/fraun/conf;1'



from .conf import *

def installVariableDatabaseNativeTool(*args, **kwd):
	from .vardb import installVariableDatabaseNativeTool
	return installVariableDatabaseNativeTool(*args, **kwd)


# Among other modules, SQLObject relies on datetime, which relies on thread-
# sensitive module imports that can fail because of race conditions.  The
# logging system uses datetime entities through sqlobject in another thread,
# and because sqlobject (as a third party library) relies on datetime, I'm
# performing the pre-import here.  It may need to be moved into the parent
# package initialization code.  The reason a pre-import should work is because
# the dynamic load won't race if the module is already in the system mapping.

try: import _strptime
except Exception as e:
	print('[_strptime pre-import] %s: %s' % (e.__class__.__name__, e))
